# modified version of https://stackoverflow.com/a/69816735
def cloudwatch_url(slug, cluster_name, region):
    def escape(s):
        for c in s:
            if c.isalpha() or c.isdigit() or c in ["-", "."]:
                continue
            c_hex = "*{0:02x}".format(ord(c))
            s = s.replace(c, c_hex)
        return s

    def gen_log_insights_url(params):
        S1 = "$257E"
        S2 = "$2528"
        S3 = "$2527"
        S4 = "$2529"

        res = f"{S1}{S2}"
        for k in params:
            value = params[k]
            if isinstance(value, str):
                value = escape(value)
            elif isinstance(value, list):
                for i in range(len(value)):
                    value[i] = escape(value[i])
            prefix = S1 if list(params.items())[0][0] != k else ""
            suffix = f"{S1}{S3}"
            if isinstance(value, list):
                value = "".join([f"{S1}{S3}{n}" for n in value])
                suffix = f"{S1}{S2}"
            elif isinstance(value, int) or isinstance(value, bool):
                value = str(value).lower()
                suffix = S1

            res += f"{prefix}{k}{suffix}{value}"
        res += f"{S4}{S4}"
        QUERY = f"logsV2:logs-insights$3Ftab$3Dlogs$26queryDetail$3D{res}"
        return f"https://{region}.console.aws.amazon.com/cloudwatch/home?region={region}#{QUERY}"

    editorString = (
        f"fields @timestamp, @message\n| sort @timestamp desc\n| "
        f"filter @logStream like /^{cluster_name}/\n| limit 200"
    )

    params = {
        "end": 0,
        "start": -60 * 60 * 24,  # 1 day ago
        "unit": "seconds",
        "timeType": "RELATIVE",  # "ABSOLUTE",  # OR RELATIVE and end = 0 and start is negative  seconds
        "tz": "Local",  # OR "UTC"
        "editorString": editorString,
        "source": [slug],
    }
    return gen_log_insights_url(params)



import json
import time
import traceback
from typing import Optional, Tuple, Union

import boto3
import botocore.exceptions
import click
import dask.config
import jsondiff
from rich import print
from rich.panel import Panel
from rich.prompt import Confirm, IntPrompt

import coiled

from ...auth import get_local_user
from ..utils import CONTEXT_SETTINGS
from .util import setup_failure

WIDTH = 90


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--region",
    default=None,
    help=(
        "AWS region to use for Coiled. By default, this will use the "
        "default region configured in your AWS profile if one is configured."
    ),
)
@click.option(
    "--profile",
    default=None,
    envvar="AWS_PROFILE",
    help=(
        "AWS profile to use from your local AWS credentials file. "
        "By default, uses your `[default]` profile if one is configured."
    ),
)
@click.option(
    "--account",
    default=None,
    help="Coiled account that will be configured to use the AWS account. By default, uses your default Coiled account.",
)
@click.option(
    "--iam-user",
    default="coiled",
    help="IAM User to create in your AWS account",
)
@click.option(
    "--setup-policy",
    default=None,
    help="Non-default name for the setup IAM Policy, default `{iam-user}-setup`",
)
@click.option(
    "--ongoing-policy",
    default=None,
    help="Non-default name for the ongoing IAM Policy, default `{iam-user}-ongoing`",
)
@click.option(
    "--update-policies",
    default=False,
    is_flag=True,
    help="Only update existing IAM Policies",
)
@click.option(
    "--update-instance-policy",
    default=False,
    is_flag=True,
    help="Update instance policy (not for regular use)",
)
@click.option(
    "--cloudshell-link",
    default=None,
    is_flag=True,
    help="Don't do setup, give instructions for setup using CloudShell",
)
@click.option(
    "--manual-final-setup",
    default=False,
    is_flag=True,
    help="Don't automatically send credentials to Coiled, finish setup manually in the web UI",
)
@click.option(
    "--quotas",
    default=False,
    is_flag=True,
    help="Check and potentially request AWS quota increases for common instance types",
)
@click.option(
    "-y",
    "--yes",
    default=False,
    is_flag=True,
    help="Don't prompt for confirmation, just do it!",
)
@click.option(
    "--accept-delete",
    default=False,
    is_flag=True,
    hidden=True,
    help="use with --yes if you also want to accept deletes",
)
@click.option(
    "--use-legacy-setup/--no-legacy-setup",
    default=False,
    is_flag=True,
    hidden=True,
    help="Use new setup backend",
)
def aws_setup(
    region: Optional[str],
    profile: Optional[str],
    account: Optional[str],
    iam_user: str,
    setup_policy: Optional[str],
    ongoing_policy: Optional[str],
    update_policies: bool,
    update_instance_policy: bool,
    cloudshell_link: Optional[bool],
    manual_final_setup: bool,
    quotas: bool,
    yes: bool,
    accept_delete: bool,
    use_legacy_setup: bool,
):
    accept_delete = yes and accept_delete  # (only accept_delete if --yes is also specified)
    if not do_setup(
        aws_profile=profile,
        slug=iam_user,
        setup_name=setup_policy,
        ongoing_name=ongoing_policy,
        just_update_policies=update_policies,
        just_update_instance_policy=update_instance_policy,
        cloudshell_link=cloudshell_link,
        region=region,
        manual_final_setup=manual_final_setup,
        quotas=quotas,
        yes=yes,
        accept_delete=accept_delete,
        use_shim=not use_legacy_setup,
        coiled_account=account,
    ):
        pass
        # print("[red]The setup process didn't finish.[/red]")


DEFAULT_REGION = "us-east-1"

TAGS = [{"Key": "owner", "Value": "coiled"}]

PROMPTS = {
    "initial": "Proceed with IAM setup?",
    "replace_access_key": (
        "Too many access keys already exist for user "
        "[green]{user_name}[/green]. "
        "\nDelete key [green]{key_id}[/green] and create a new key?"
    ),
    "request_quotas": "Would you like to make any AWS quota increase requests?",
}

SCRIPT_REQUIRED_IAM = [
    "iam:GetPolicy",
    "iam:ListAccessKeys",
    "iam:CreateUser",
    "iam:TagUser",
    "iam:CreatePolicy",
    "iam:CreatePolicyVersion",
    "iam:AttachUserPolicy",
    "iam:GetPolicyVersion",
    "iam:CreateAccessKey",
    "iam:DeleteAccessKey",
]

setup_doc = """{
   "Statement": [
      {
         "Sid": "Setup",
         "Effect": "Allow",
         "Resource": "*",
         "Action": [
            "ec2:AssociateRouteTable",
            "ec2:AttachInternetGateway",
            "ec2:CreateInternetGateway",
            "ec2:CreateRoute",
            "ec2:CreateRouteTable",
            "ec2:CreateSubnet",
            "ec2:CreateVpc",
            "ec2:DeleteInternetGateway",
            "ec2:DeleteRoute",
            "ec2:DeleteRouteTable",
            "ec2:DeleteSubnet",
            "ec2:DeleteVpc",
            "ec2:DescribeInternetGateways",
            "ec2:DetachInternetGateway",
            "ec2:DisassociateRouteTable",
            "ec2:ModifySubnetAttribute",
            "ec2:ModifyVpcAttribute",
            "iam:AddRoleToInstanceProfile",
            "iam:AttachRolePolicy",
            "iam:CreateRole",
            "iam:CreatePolicy",
            "iam:CreateServiceLinkedRole",
            "iam:CreateInstanceProfile",
            "iam:DeleteRole",
            "iam:ListPolicies",
            "iam:ListInstanceProfiles",
            "iam:ListAttachedRolePolicies",
            "iam:TagInstanceProfile",
            "iam:TagPolicy",
            "iam:TagRole"
         ]
      }
   ],
   "Version": "2012-10-17"
}"""

ongoing_doc = """{"Statement": [
      {
         "Sid": "Ongoing",
         "Effect": "Allow",
         "Resource": "*",
         "Action": [
            "ec2:AuthorizeSecurityGroupIngress",
            "ec2:CreateFleet",
            "ec2:CreateLaunchTemplate",
            "ec2:CreateLaunchTemplateVersion",
            "ec2:CreateRoute",
            "ec2:CreateSecurityGroup",
            "ec2:CreateTags",
            "ec2:DeleteFleets",
            "ec2:DeleteLaunchTemplate",
            "ec2:DeleteLaunchTemplateVersions",
            "ec2:DeleteSecurityGroup",
            "ec2:DescribeAvailabilityZones",
            "ec2:DescribeImages",
            "ec2:DescribeInstances",
            "ec2:DescribeInstanceTypeOfferings",
            "ec2:DescribeInstanceTypes",
            "ec2:DescribeInternetGateways",
            "ec2:DescribeLaunchTemplates",
            "ec2:DescribeNatGateways",
            "ec2:DescribeNetworkInterfaces",
            "ec2:DescribeRegions",
            "ec2:DescribeRouteTables",
            "ec2:DescribeSecurityGroups",
            "ec2:DescribeSubnets",
            "ec2:DescribeVpcPeeringConnections",
            "ec2:DescribeVpcs",
            "ec2:RunInstances",
            "ec2:TerminateInstances",
            "ecr:BatchCheckLayerAvailability",
            "ecr:BatchGetImage",
            "ecr:CompleteLayerUpload",
            "ecr:CreateRepository",
            "ecr:DescribeImages",
            "ecr:DescribeRepositories",
            "ecr:GetAuthorizationToken",
            "ecr:GetDownloadUrlForLayer",
            "ecr:GetRepositoryPolicy",
            "ecr:InitiateLayerUpload",
            "ecr:ListImages",
            "ecr:PutImage",
            "ecr:UploadLayerPart",
            "ecr:TagResource",
            "iam:GetInstanceProfile",
            "iam:GetRole",
            "iam:ListAttachedRolePolicies",
            "iam:ListInstanceProfiles",
            "iam:ListPolicies",
            "iam:PassRole",
            "iam:TagRole",
            "logs:CreateLogGroup",
            "logs:CreateLogStream",
            "logs:DescribeLogGroups",
            "logs:DescribeLogStreams",
            "logs:GetLogEvents",
            "logs:FilterLogEvents",
            "logs:PutLogEvents",
            "logs:PutRetentionPolicy",
            "logs:TagLogGroup",
            "logs:TagResource",
            "sts:GetCallerIdentity"
         ]
      }
   ],
   "Version": "2012-10-17"
}"""


def create_user(iam, name):
    arn = None
    try:
        r = iam.create_user(UserName=name, Tags=TAGS)
        arn = r["User"]["Arn"]
        coiled.add_interaction(action="CreateUser", success=True, arn=arn)
    except iam.exceptions.EntityAlreadyExistsException:
        print(f"user [green]{name}[/green] already exists")
    return arn


def create_or_update_policy(iam, aws_account, name, doc):
    try:
        r = iam.create_policy(PolicyName=name, PolicyDocument=doc)
        arn = r["Policy"]["Arn"]
        coiled.add_interaction(action="CreatePolicy", success=True, arn=arn)
    except iam.exceptions.EntityAlreadyExistsException:
        arn = f"arn:aws:iam::{aws_account}:policy/{name}"
        update_policy(iam, aws_account, name, doc)

    return arn


def get_policy_diff(iam, aws_account, name, doc):
    existing_arn, changes = None, None

    arn = f"arn:aws:iam::{aws_account}:policy/{name}"

    try:
        policy = iam.get_policy(PolicyArn=arn)
        existing_arn = arn
    except iam.exceptions.ClientError as e:
        error_code = e.response["Error"].get("Code")

        if error_code == "NoSuchEntity":
            return existing_arn, changes
        else:
            raise

    policy_version = iam.get_policy_version(PolicyArn=arn, VersionId=policy["Policy"]["DefaultVersionId"])
    existing_doc = policy_version["PolicyVersion"]["Document"]

    doc_diff = jsondiff.diff(existing_doc, json.loads(doc), syntax="symmetric")

    if doc_diff:
        changes = doc_diff

    return existing_arn, changes


def show_policy_diff(doc_diff) -> str:
    diff_lines = []

    if doc_diff:
        if "Statement" in doc_diff and len(doc_diff) == 1:
            if 0 in doc_diff["Statement"] and len(doc_diff["Statement"]) == 1:
                if "Action" in doc_diff["Statement"][0] and len(doc_diff["Statement"][0]) == 1:
                    action_changes = doc_diff["Statement"][0]["Action"]
                    line_changes = []

                    if jsondiff.symbols.insert in action_changes:
                        line_changes.extend([(i, "+", n) for (i, n) in action_changes[jsondiff.symbols.insert]])
                    if jsondiff.symbols.delete in action_changes:
                        line_changes.extend([(i, "-", n) for (i, n) in action_changes[jsondiff.symbols.delete]])

                    for i, change, line in sorted(line_changes):
                        line_color = "red" if change == "-" else "green"
                        diff_lines.append(f" [{line_color}][bold]{change}[/bold] ({i}) {line}")

                    return "\n".join(diff_lines)

        # if we couldn't print the short version, just print the raw diff object
        return f"{doc_diff}"
    else:
        return ""


def update_policy(iam, aws_account, name, doc):
    policy_arn = f"arn:aws:iam::{aws_account}:policy/{name}"
    try:
        r = iam.create_policy_version(PolicyArn=policy_arn, PolicyDocument=doc, SetAsDefault=True)
        coiled.add_interaction(action="CreatePolicyVersion", success=True, arn=policy_arn)
        new_version = r["PolicyVersion"]["VersionId"]
        print(f"Updated Policy [green]{policy_arn}[/green] is [bold]{new_version}[/bold]")
        print()
    except iam.exceptions.LimitExceededException:
        # this is Coiled-specific policy so should be fine to delete old version
        existing_policies = iam.list_policy_versions(PolicyArn=policy_arn)
        to_delete = [
            version["VersionId"] for version in existing_policies["Versions"] if not version["IsDefaultVersion"]
        ][-1]
        print(f"Policy {name} has too many existing versions, deleting {to_delete}")
        iam.delete_policy_version(PolicyArn=policy_arn, VersionId=to_delete)
        update_policy(iam, aws_account, name, doc)
    except Exception as e:
        coiled.add_interaction(
            action="CreatePolicyVersion",
            success=False,
            arn=policy_arn,
            error_message=str(e),
        )
        print("[red]Unable to update existing policy[/red]:")
        print(f"  [red]{e}[/red]")
        print()


def attach_policy(iam, user, policy_arn):
    # idempotent
    iam.attach_user_policy(UserName=user, PolicyArn=policy_arn)
    coiled.add_interaction(action="AttachUserPolicy", success=True, arn=policy_arn)


def make_access_key(iam, user, *, retry=0, accept_delete: bool = False) -> Tuple[Optional[str], Optional[str]]:
    try:
        r = iam.create_access_key(UserName=user)
        coiled.add_interaction(action="CreateAccessKey", success=True, user=user)

        key_id = r["AccessKey"]["AccessKeyId"]
        key_secret = r["AccessKey"]["SecretAccessKey"]

        return key_id, key_secret

    except iam.exceptions.LimitExceededException:
        coiled.add_interaction(
            action="CreateAccessKey",
            success=False,
            user=user,
            error_message="LimitExceededException",
        )
        if retry:
            print("[red]already retried, giving up[/red]")
            return None, None

        # FIXME let user select which key to delete
        # if Confirm.ask(PROMPTS["replace_access_key"].format(user_name=user, key_id=key_id), default=True):
        if delete_access_key(iam, user, accept_delete=accept_delete):
            return make_access_key(iam, user, retry=retry + 1, accept_delete=accept_delete)

    return None, None


def get_user_access_keys(iam, user):
    try:
        r = iam.list_access_keys(UserName=user)
        coiled.add_interaction(action="ListAccessKeys", success=True, user=user)
        return True, r["AccessKeyMetadata"]
    except iam.exceptions.ClientError as e:
        error_code = e.response["Error"].get("Code")

        if error_code == "NoSuchEntity":
            coiled.add_interaction(
                action="ListAccessKeys",
                success=False,
                user=user,
                error_message=str(e),
            )
            return False, []
        else:
            raise


def delete_access_key(iam, user, accept_delete: bool = False):
    r = iam.list_access_keys(UserName=user)
    key_id = r["AccessKeyMetadata"][0]["AccessKeyId"]

    # NOTE: --yes doesn't apply to this since currently
    print()
    if accept_delete or Confirm.ask(
        PROMPTS["replace_access_key"].format(user_name=user, key_id=key_id),
        default=True,
    ):
        coiled.add_interaction(action="prompt:DeleteAccessKey", success=True, user=user)
        iam.delete_access_key(UserName=user, AccessKeyId=key_id)
        coiled.add_interaction(action="DeleteAccessKey", success=True, user=user, key_id=key_id)
        print(f"Deleted access key [green]{key_id}[/green]")
        return True
    else:
        coiled.add_interaction(action="prompt:DeleteAccessKey", success=False, user=user)

    return False


def wait_till_key_works(iam, key, secret):
    print("")
    print("Waiting until access key is ready", end="")
    t0 = time.time()

    while t0 + 10 > time.time():
        try:
            coiled.utils.verify_aws_credentials(key, secret)
            # a little extra wait here also seems to help, otherwise we *still* yet error sometimes when trying to use
            print(".", end="")
            time.sleep(1)
            print()

            return True
        except iam.exceptions.ClientError as e:
            error_code = e.response["Error"].get("Code")
            if error_code == "InvalidClientTokenId":
                print(".", end="")
                time.sleep(1)
                continue

    print("\nAccess key is still not ready. Please manually set up your Coiled account with AWS.")
    return False


def do_intro(sts, iam, region, coiled_account):
    introduction = """
[bold]Introduction[/bold]

This uses your AWS credentials to set up Coiled.

This will do the following ...
1. Create limited IAM roles and grant them to Coiled
2. Check and expand your AWS quota if needed
3. Create initial resources to deploy clusters

This will [bold]not[/bold] ...
1. Create resources that cost money
2. Grant Coiled access to your data
""".strip()
    print(Panel(introduction, width=WIDTH))

    try:
        # get_caller_identity doesn't require any specific IAM rights
        identity = sts.get_caller_identity()
        aws_account = identity.get("Account")
        identity_as = identity.get("Arn").split(":")[-1]

        coiled.add_interaction(
            action="GetCallerIdentity",
            success=True,
            account=aws_account,
            identity_as=identity_as,
        )

        try:
            r = iam.list_account_aliases()
            account_alias = r.get("AccountAliases", [])[0]
            alias_string = f" ({account_alias})"
            coiled.add_interaction(
                action="ListAccountAliases",
                success=True,
                account=aws_account,
                alias=alias_string,
            )
        except Exception:
            # doesn't matter much if we can't get account alias
            alias_string = ""

        account_desc = f"""
[bold]Target AWS Account[/bold]

Current credentials:\t[green]{identity_as}[/green]
Proposed region :\t[green]{region}[/green]
Proposed account:\t[green]{aws_account}[/green]{alias_string}

If this is incorrect then stop and select a different ...
  profile using the [green]coiled setup aws [bold]--profile[/bold][/green] argument
  region using the [green]coiled setup aws [bold]--region[/bold][/green] argument
""".strip()

        print(Panel(account_desc, width=WIDTH))

        return aws_account
    except botocore.exceptions.NoCredentialsError:
        coiled.add_interaction(
            action="GetCallerIdentity",
            success=False,
            error_message="NoCredentialsError",
        )

        print(
            "[red]Missing:[/red] You don't have local AWS credentials.\n"
            "That's ok, you can run setup from AWS CloudShell."
        )
        setup_failure("Getting local aws credentials failed", backend="aws")
        print()
        # is aws cli installed?
        # get_aws_cli_version = subprocess.run(["aws", "--version"], capture_output=True)
        # has_aws_cli = get_aws_cli_version.returncode == 0

        show_cloudshell_instructions(region, coiled_account=coiled_account)

    except Exception as e:
        coiled.add_interaction(action="GetCallerIdentity", success=False, error_message=str(e))
        setup_failure(f"Getting local aws credentials failed {str(e)}", backend="aws")
        print("Error determining your AWS account:")
        print(f"    [red]{e}[/red]")
        print()

        show_cloudshell_instructions(region, coiled_account=coiled_account)

    return None


def show_cloudshell_instructions(region, coiled_account):
    # explain cloudshell
    server_arg = (
        f"--server {dask.config.get('coiled.server', coiled.utils.COILED_SERVER)} \\ "
        if dask.config.get("coiled.server", coiled.utils.COILED_SERVER) != coiled.utils.COILED_SERVER
        else ""
    )

    token = dask.config.get("coiled.token")
    a, b, c = token[:24], token[24:48], token[48:]
    token_args = [
        f"  --token {a} \\",
        f"  --token {b} \\",
        f"  --token {c} && \\",
    ]
    region_arg = f" --region {region}" if region else ""
    account_arg = f"--account {coiled_account} " if coiled_account else ""

    cli_lines = [
        "pip3 install coiled && \\ ",
        f"coiled login {account_arg}\\ ",
    ]
    if server_arg:
        cli_lines.append(server_arg)
    cli_lines.extend(
        [
            *token_args,
            f"coiled setup aws{region_arg}",
        ]
    )
    cli_command = "\n  ".join(cli_lines)

    instruction_text = (
        "Run setup from AWS CloudShell with the following steps:\n\n"
        "1. Go to [link]https://console.aws.amazon.com/cloudshell[/link]\n"
        "2. Sign in to your AWS account\n"
        "   (if you usually switch role or profile, you should do this)\n"
        "3. Run the following command in CloudShell:\n\n"
        f"  [green]{cli_command}[/green]\n"
    )

    # box might be nice but would make copying the command much worse
    print(instruction_text)


def do_coiled_setup(iam, key, secret, region, coiled_account, yes, use_shim) -> bool:
    success = False
    backend_failure = False

    if key and secret:
        coiled.add_interaction(action="prompt:CoiledSetup", success=True)
        # wait on this for a bit till we don't get InvalidClientTokenId error
        if wait_till_key_works(iam, key, secret):
            print("Setting up Coiled to use your AWS account...")
            try:
                coiled.set_backend_options(
                    account=coiled_account,
                    backend="aws",
                    aws_access_key_id=key,
                    aws_secret_access_key=secret,
                    aws_region=region,
                    use_shim=use_shim,
                )
                success = True
                coiled.add_interaction(action="CoiledSetup", success=True)
            except Exception as e:
                backend_failure = True
                error_message = str(e)
                coiled.add_interaction(action="CoiledSetup", success=False, error_message=error_message)
                print()
                print("[red]There was an error setting up Coiled to use your AWS account:\n")
                print(error_message)
        else:
            coiled.add_interaction(action="prompt:CoiledSetup", success=False)

        message = """
[bold]Setup complete 🎉[/bold]

What's next?

  Run a command line application in the cloud with:

    $ [bold]coiled run echo 'Hello, world'[/bold]

  Or create a Dask cluster with:

    $ ipython

    [bold]import coiled
    cluster = coiled.Cluster(
        n_workers=10,
    )
    client = cluster.get_client()[/bold]

  For more examples see [link]https://docs.coiled.io/user_guide/examples/index.html[/link]
""".strip()

        print()
        print()
        print(message)

        if not success and not backend_failure:
            show_manual_setup(key, secret, coiled_account=coiled_account)

    return success


def do_full_setup(
    session,
    iam,
    user_name,
    aws_account,
    region,
    setup_name,
    ongoing_name,
    *,
    coiled_account,
    manual_final_setup,
    yes,
    accept_delete,
    use_shim,
) -> bool:
    # Check for existing user
    user_exists, user_keys = get_user_access_keys(iam, user_name)

    # Check for existing policies
    setup_arn, setup_diff = get_policy_diff(iam, aws_account, setup_name, setup_doc)
    ongoing_arn, ongoing_diff = get_policy_diff(iam, aws_account, ongoing_name, ongoing_doc)

    resource_strings = ["[bold]Proposed AWS Account Changes[/bold]\n"]

    if not user_exists:
        resource_strings.append(f"Create IAM User:\t[green]{user_name}[/green]")
        resource_strings.append("  and create Access Key for this new IAM User")
    else:
        resource_strings.append(f"Create Access Key for [bold]existing[/bold] IAM User:\t[green]{user_name}[/green]")

        existing_key_message = f"  This IAM User already has {len(user_keys)} access keys."
        if len(user_keys) > 1:
            existing_key_message = (
                f"  This IAM User already has {len(user_keys)} access keys. \n"
                "  We'll need to [bold]delete[/bold] an existing access key to create a new access key."
            )
        resource_strings.append(existing_key_message)
        resource_strings.append("")
        resource_strings.append(
            "If you didn't want to configure Coiled with this existing IAM User, "
            "stop now and specify a different IAM User name with the "
            "[green]coiled setup aws [bold]--iam-user[/bold][/green] argument."
        )

    if not setup_arn:
        resource_strings.append("")
        resource_strings.append(f"Create IAM Policy:\t[green]{setup_name}[/green]")
        resource_strings.append(f"  and attach to IAM User [green]{user_name}[/green]")
    if setup_diff:
        resource_strings.append(f"Update IAM Policy:\t[green]{setup_name}[/green]")

    if not ongoing_arn:
        resource_strings.append("")
        resource_strings.append(f"Create IAM Policy:\t[green]{ongoing_name}[/green]")
        resource_strings.append(f"  and attach to IAM User [green]{user_name}[/green]")
    if ongoing_diff:
        resource_strings.append(f"Update IAM Policy:\t[green]{ongoing_name}[/green]")

    if setup_diff:
        resource_strings.append("")
        resource_strings.append(f"Proposed changes to the existing [green]{setup_name}[/green] IAM Policy:")
        resource_strings.append(show_policy_diff(setup_diff))

    if ongoing_diff:
        resource_strings.append("")
        resource_strings.append(f"Proposed changes to the existing [green]{ongoing_name}[/green] IAM Policy:")
        resource_strings.append(show_policy_diff(ongoing_diff))

    if not setup_arn or not ongoing_arn:
        resource_strings.append(
            "\nDocumentation for IAM Policies at "
            "[link]https://docs.coiled.io/user_guide/aws_configure.html#create-iam-policies[/link]"
        )

    resource_desc = "\n".join(resource_strings)
    print(Panel(resource_desc, width=WIDTH))

    if not yes and not Confirm.ask(PROMPTS["initial"], default=True):
        coiled.add_interaction(action="prompt:Setup_AWS", success=False)
        return False

    coiled.add_interaction(action="prompt:Setup_AWS", success=True)

    create_arns = []

    if not user_exists:
        user_arn = create_user(iam, user_name)
        if user_arn:
            create_arns.append(user_arn)

    if not setup_arn:
        setup_arn = create_or_update_policy(iam, aws_account, setup_name, setup_doc)
        create_arns.append(setup_arn)
    elif setup_diff:
        setup_arn = create_or_update_policy(iam, aws_account, setup_name, setup_doc)

    if not ongoing_arn:
        ongoing_arn = create_or_update_policy(iam, aws_account, ongoing_name, ongoing_doc)
        create_arns.append(ongoing_arn)
    elif ongoing_diff:
        ongoing_arn = create_or_update_policy(iam, aws_account, ongoing_name, ongoing_doc)

    attach_policy(iam, user_name, setup_arn)
    attach_policy(iam, user_name, ongoing_arn)

    print()
    if create_arns:
        print("The following resources were created in your AWS account:")
        for arn in create_arns:
            print(f"  {arn}")
    print(f"IAM User [green]{user_name}[/green] is now setup with IAM Policies attached.")
    print()

    check_quotas(session, region=region)

    key, secret = make_access_key(iam, user_name, accept_delete=accept_delete)

    if manual_final_setup:
        return show_manual_setup(key, secret, coiled_account)
    else:
        return do_coiled_setup(
            iam=iam, key=key, secret=secret, region=region, coiled_account=coiled_account, yes=yes, use_shim=use_shim
        )


def show_manual_setup(key, secret, coiled_account):
    if not coiled_account:
        with coiled.Cloud() as cloud:
            coiled_account = cloud.default_account

    setup_url = (
        f"{dask.config.get('coiled.server', coiled.utils.COILED_SERVER)}/"
        f"{coiled_account}/settings/setup/credentials"
    )

    print(
        Panel(
            "You've successfully configured an IAM User for Coiled "
            "in your AWS account. "
            "You can now complete your Coiled account setup by granting "
            "us access to this IAM User:\n"
            "\n"
            f"1. Go to [link]{setup_url}[/link] and select AWS.\n"
            "\n"
            # FIXME adjust instructions once there's better web UI flow
            #  ideally we'd have direct URL to relevant next step (where you'd enter access key)
            '2. Select "Browser Setup".\n'
            "\n"
            "3. Select your desired default AWS region and enter these values where prompted:\n"
            "\n"
            f"\tAWS Access Key ID:\t[green]{key}[/green]\n"
            f"\tAWS Secret Access Key:\t[green]{secret}[/green]\n"
            "\n"
            "4. Continue with the other account setup steps; you'll be able to "
            "choose non-default network or container registry settings as desired.",
            width=WIDTH,
        )
    )

    return True


def do_just_update_policies(iam, aws_account, setup_name, ongoing_name, yes) -> bool:
    # Check for existing policies
    setup_arn, setup_diff = get_policy_diff(iam, aws_account, setup_name, setup_doc)
    ongoing_arn, ongoing_diff = get_policy_diff(iam, aws_account, ongoing_name, ongoing_doc)

    if not setup_arn:
        print(f"[red]WARNING[/red]: No IAM Policy named [green]{setup_name}[/green] found.")
        print("Use `--setup-policy` to specify a different name for the existing setup policy.")
        print()
        return False

    if not ongoing_arn:
        print(f"[red]WARNING[/red]: No IAM Policy named [green]{ongoing_name}[/green] found")
        print("Use `--ongoing-policy` to specify a different name for the existing ongoing policy.")
        print()
        return False

    if setup_diff:
        print(f"Proposed changes to the existing [green]{setup_name}[/green] IAM Policy:")
        show_policy_diff(setup_diff)
        print()

    if ongoing_diff:
        print(f"Proposed changes to the existing [green]{ongoing_name}[/green] IAM Policy:")
        show_policy_diff(ongoing_diff)
        print()

    if setup_arn and ongoing_arn and not setup_diff and not ongoing_diff:
        print("Your AWS IAM Policies are up-to-date")
    elif setup_arn and not setup_diff:
        print("Your [bold]setup[/bold] IAM Policy is up-to-date")
        print("You may need to update your [bold]ongoing[/bold] IAM Policy.")
    elif ongoing_arn and not ongoing_diff:
        print("Your [bold]ongoing[/bold] IAM Policy is up-to-date")
        print("You may need to update your [bold]setup[/bold] IAM Policy.")

    elif not yes and not Confirm.ask(PROMPTS["initial"], default=True):
        return False

    if setup_diff:
        update_policy(iam, aws_account, setup_name, setup_doc)

    if ongoing_diff:
        update_policy(iam, aws_account, ongoing_name, ongoing_doc)

    return True


def get_session(aws_profile: Optional[str], region_name: Optional[str], skip_telemetry: bool = False):
    try:
        session = boto3.Session(profile_name=aws_profile, region_name=region_name)
        if not skip_telemetry:
            coiled.add_interaction(
                action="BotoSession",
                success=True,
                profile=aws_profile,
                region_name=session.region_name,
            )
        return session
    except botocore.exceptions.ProfileNotFound:
        coiled.add_interaction(
            action="BotoSession",
            success=False,
            profile=aws_profile,
            error_message="ProfileNotFound",
        )
        print()
        print(f"[red]The profile `{aws_profile}` is not configured in your local AWS credentials.")
        print(
            "If this isn't the correct AWS identity or account, you can specify a different profile "
            "from your AWS credentials file using the "
            "[green]coiled setup aws [bold]--profile[/bold][/green] argument"
        )
        setup_failure("Requested AWS profile not found", backend="aws")
        return None


def do_setup(
    slug,
    aws_profile=None,
    setup_name=None,
    ongoing_name=None,
    just_update_policies=False,
    just_update_instance_policy=False,
    region=None,
    cloudshell_link=None,
    manual_final_setup=None,
    quotas=None,
    yes=False,
    accept_delete=False,
    use_shim=False,
    coiled_account=None,
) -> bool:
    local_user = get_local_user()

    # this isn't perfect since we aren't checking if token is valid, but it's better than nothing
    is_logged_in = bool(dask.config.get("coiled.token", None))

    # We don't want to require Coiled account if just using `coiled setup aws --quotas`, thus
    # skipping telemetry for `--quotas` if user doesn't already have local Coiled token.
    skip_telemetry = bool(quotas and not is_logged_in)

    if not is_logged_in and not quotas:
        print("To set up your Coiled account, you'll need to authorize this computer to access your Coiled account.")

    if not skip_telemetry:
        # This will trigger auth flow if necessary.
        try:
            coiled.add_interaction(
                action="CliSetupAwsQuotas" if quotas else "CliSetupAws",
                success=True,
                local_user=local_user,
                # use keys that match the cli args
                profile=aws_profile,
                iam_user=slug,
                setup_policy=setup_name,
                ongoing_policy=ongoing_name,
                update_policies=just_update_policies,
                update_instance_policy=just_update_instance_policy,
                region=region,
                cloudshell_link=cloudshell_link,
                manual_final_setup=manual_final_setup,
                yes=yes,
                accept_delete=accept_delete,
                coiled_account=coiled_account,
                quotas=quotas,
            )
        except KeyboardInterrupt:
            if not bool(dask.config.get("coiled.token", None)):
                print("[red]This computer doesn't have access to your Coiled account so setup cannot continue.")
                print("Try running [green]coiled login[/green] to authorize access to your Coiled account.")
                return False
        except ValueError as e:
            if "Authorization failed" in str(e):
                print("[red]This computer doesn't have access to your Coiled account so setup cannot continue.")
                print("Try running [green]coiled login[/green] to authorize access to your Coiled account.")
                return False
            else:
                raise

    if cloudshell_link:
        show_cloudshell_instructions(region, coiled_account=coiled_account)
        return False

    try:
        session = get_session(aws_profile, region_name=region, skip_telemetry=skip_telemetry)
        if session is None:
            print()
            print(
                "If you don't have the AWS credentials locally, another option is to use CloudShell in the AWS Console."
            )
            show_cloudshell_instructions(region=region, coiled_account=coiled_account)
            return False

        region = region or session.region_name or DEFAULT_REGION

        if quotas:
            try:
                # check quotas and give option to request increases
                check_quotas(session, region=region, just_quota_flow=True, skip_telemetry=skip_telemetry)
            except KeyboardInterrupt:
                pass
            except Exception as e:
                raise e
            return False

        try:
            iam = session.client("iam")
        except Exception as e:
            print()
            print("Something went wrong when trying to use your local AWS credentials.")
            print()
            print(f"[red][bold]{type(e).__name__}[/bold]: {e}[/red]")
            print()
            show_cloudshell_instructions(region, coiled_account=coiled_account)
            return False

        sts = session.client("sts")

        user_name = slug
        setup_name = setup_name or f"{slug}-setup"
        ongoing_name = ongoing_name or f"{slug}-ongoing"

        try:
            aws_account = do_intro(sts, iam, region=region, coiled_account=coiled_account)
            if not aws_account:
                return False

            if just_update_instance_policy:
                return update_instance_profile_policy(iam, aws_account, yes=yes)
            elif just_update_policies:
                return do_just_update_policies(iam, aws_account, setup_name, ongoing_name, yes=yes)
            else:
                return do_full_setup(
                    session,
                    iam,
                    user_name,
                    aws_account,
                    region,
                    setup_name,
                    ongoing_name,
                    coiled_account=coiled_account,
                    manual_final_setup=manual_final_setup,
                    yes=yes,
                    accept_delete=accept_delete,
                    use_shim=use_shim,
                )

        except iam.exceptions.ClientError as e:
            error_code = e.response["Error"].get("Code")
            error_msg = e.response["Error"].get("Message")
            error_op = e.operation_name

            coiled.add_interaction(action=error_op, success=False, error_message=error_msg)

            if "assumed-role/AmazonSageMaker-ExecutionRole" in error_msg:
                print()
                print("It appears that you're trying to set up Coiled from inside Amazon SageMaker.")
                print(
                    "SageMaker has restricted permissions on your AWS account. Although you [bold]can use[/bold] "
                    "Coiled from a SageMaker notebook, you [bold]cannot set up[/bold] Coiled from SageMaker."
                )
                print()
                setup_failure("Inside sagemaker", backend="aws")
                show_cloudshell_instructions(region, coiled_account=coiled_account)
                return False

            elif "AccessDenied" in str(e):
                print()
                print(f"Insufficient permissions to [green]{error_op}[/green] using current AWS profile/user.")
                print("You may want to try with a different AWS profile that has different permissions.")
                print()
                print(f"[red]{error_msg}[/red]")
                print()
                print("To run this setup script you'll need the following IAM permissions:")
                for permission in SCRIPT_REQUIRED_IAM:
                    print(f"- {permission}")
                print(
                    "If you don't have access to an AWS profile with these permissions, you may need to ask "
                    "someone with administrative access to your AWS account to help you create the IAM User "
                    "and IAM Policies described in our documentation: "
                    "[link]https://docs.coiled.io/user_guide/aws/manual.html[/link]"
                )
                setup_failure(
                    f"Permission error during for {error_op}. {error_msg}",
                    backend="aws",
                )
                return False
            else:
                print()
                print(f"Something went wrong when trying to [green]{error_op}[/green].")
                print()
                print(f"[red][bold]{error_code}[/bold]: {error_msg}[/red]")
                print()
                setup_failure(
                    f"Error trying {error_op}. {error_msg}",
                    backend="aws",
                )
                return False

    except KeyboardInterrupt as e:
        tb = "\n".join(traceback.format_tb(e.__traceback__))
        coiled.add_interaction(action="KeyboardInterrupt", success=False, error_message=tb)
        raise

    # catch all so we make sure all errors are tracked
    except Exception as e:
        msg = traceback.format_exc()
        coiled.add_interaction(action="Unknown", success=False, error_message=msg)
        # TODO better generic error handling
        print()
        print("[red][bold]ERROR[/bold][/red] Something unexpected happened:")
        print(f"    [red]{e}")
        print("Please reach out to Coiled Support at support@coiled.io if you need help with this issue.")
        setup_failure(
            f"Unhandled exception {msg}",
            backend="aws",
        )
        return False


def check_local_aws_creds():
    session = boto3.Session()
    sts = session.client("sts")
    try:
        # good call to try, since this doesn't require any IAM permissions
        sts.get_caller_identity()
        return True
    except botocore.exceptions.NoCredentialsError:
        return False


def update_instance_profile_policy(iam, aws_account, yes):
    policy_name = "CoiledInstancePolicy"
    policy_doc = """{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "CoiledEC2Policy",
            "Effect": "Allow",
            "Action": [
                "logs:CreateLogGroup",
                "logs:CreateLogStream",
                "logs:PutLogEvents",
                "cloudwatch:PutMetricData",
                "aps:RemoteWrite"
            ],
            "Resource": "*"
        }
    ]
}"""

    # Check for existing policy
    policy_arn, policy_diff = get_policy_diff(iam, aws_account, policy_name, policy_doc)

    if not policy_arn:
        print(f"[red]WARNING[/red]: No IAM Policy named [green]{policy_name}[/green] found.")
        print()

    if policy_diff:
        print(f"Proposed changes to the existing [green]{policy_name}[/green] IAM Policy:")
        show_policy_diff(policy_diff)
        print()

    if policy_arn and not policy_diff:
        print(f"Your AWS policy document for {policy_name} is up-to-date")

    elif not yes and not Confirm.ask(f"Update {policy_name} policy document?", default=True):
        return False

    if policy_diff:
        update_policy(iam, aws_account, policy_name, policy_doc)

    return True


def check_quotas(session, region: str, just_quota_flow: bool = False, skip_telemetry: bool = False):
    quota_client = session.client("service-quotas", region_name=region)

    quotas = [
        {
            "class": "Standard On-Demand",
            "name": "Running On-Demand Standard (A, C, D, H, I, M, R, T, Z) instances",
            "code": "L-1216C47A",
            "low": 44,
        },
        {
            "class": "Standard Spot",
            "name": "All Standard (A, C, D, H, I, M, R, T, Z) Spot Instance Requests",
            "code": "L-34B43A08",
            "low": 44,
        },
        {
            "class": "G4dn (NVIDIA T4 GPU) On-Demand",
            "name": "Running On-Demand G and VT instances",
            "code": "L-DB2E81BA",
            "low": -1,
        },
        {
            "class": "G4dn (NVIDIA T4 GPU) Spot",
            "name": "All G and VT Spot Instance Requests",
            "code": "L-3819A6DF",
            "low": -1,
        },
        {
            "class": "P (NVIDIA V100/A100 GPU) On-Demand",
            "name": "Running On-Demand P instances",
            "code": "L-417A185B",
            "low": -1,
        },
        {
            "class": "P (NVIDIA V100/A100 GPU) Spot",
            "name": "All P Spot Instance Requests",
            "code": "L-7212CCBC",
            "low": -1,
        },
    ]

    quota_lines = []

    for quota in quotas:
        quota["link"] = get_quota_link(region, quota["code"])

    error_message = None
    got_quotas = False
    low_quotas = []
    try:
        for quota in quotas:
            quota["value"] = get_quota_value(quota_client, quota["code"])
            if quota.get("low") is not None and quota["value"] <= quota["low"]:
                quota["is_low"] = True
                low_quotas.append(quota)
            got_quotas = True
    except quota_client.exceptions.AccessDeniedException:
        error_message = (
            "[red]Your active AWS account doesn't have permission to read service quotas.[/red]\n\n"
            "You can check the quotas in the AWS Console using the links below."
        )
    except Exception as e:
        error_message = (
            "An unexpected error occurred while trying to read your AWS service quotas:\n\n"
            f"[red]{e}[/red]\n\n"
            "You can check the quotas in the AWS Console using the links below."
        )

    for quota in quotas:
        quota_label = quota["class"]
        if quota.get("value") is not None:
            quota_value_text = f"{quota.get('value'):>6} vCPU"
            if quota.get("is_low"):
                quota_value_text = f"{quota_value_text} (this quota is low, you may wish to request increase)"
        else:
            quota_value_text = f"[link]{quota['link']}[/link]"

        quota_lines.append(f"{quota_label:<36}{quota_value_text}")

    quota_text = "\n".join(quota_lines)

    quota_text = (
        "[bold]Current AWS Quotas[/bold]\n\n"
        f"{quota_text}\n\n"
        "[bold]Standard[/bold] includes:\n"
        "    general purpose [bold]M[/bold] and [bold]T[/bold] families (e.g., M6i, T3),\n"
        "    compute optimized [bold]C[/bold] families (e.g., C6i),\n"
        "    memory optimized [bold]R[/bold] families (e.g., R6i).\n"
        "\n[bold]GPU[/bold] instances have a separate quotas based on GPU type."
    )

    if got_quotas and not just_quota_flow:
        example_cost = "2.46" if region == "us-west-1" else "2.11"
        example_region = region if region in ("us-east-1", "us-east-2", "us-west-1", "us-west-2") else "us-east-1"

        quota_text = (
            f"{quota_text}\n\n"
            "[bold]Example Usage:[/bold]\n"
            "10 VM cluster with m6i.xlarge instances would have 40 vCPUs.\n"
            f"AWS compute cost would be ${example_cost}/hr "
            f"for on-demand instances in {example_region}."
        )

    if error_message:
        quota_text = f"{error_message}\n\n{quota_text}"

    print(Panel(quota_text, width=WIDTH))

    if got_quotas:
        if just_quota_flow:
            # when running explicit quota check (`--quotas`), prompt about each quota
            do_quota_increases(quota_client, region, quotas, skip_telemetry=skip_telemetry)
        elif low_quotas:
            print(
                "Some of your quotas are low. If you'd like to request any increases for these quotas, "
                "we can prompt you for the desired quota value and attempt to submit this request to AWS.\n"
            )
            if Confirm.ask(PROMPTS["request_quotas"], default=True):
                # when running normal setup flow, just prompt about low quotas
                # note that we never consider GPU quotas low ("low" threshold is set to -1)
                do_quota_increases(quota_client, region, low_quotas, skip_telemetry=skip_telemetry)


def get_quota_value(quota_client, quota_code: str) -> int:
    quota = quota_client.get_service_quota(
        ServiceCode="ec2",
        QuotaCode=quota_code,
    )
    return int(quota["Quota"]["Value"])


def do_quota_increases(quota_client, region: str, quotas: list, skip_telemetry: bool):
    for quota in quotas:
        prompt = (
            f"[bold]{quota['class']}[/bold] ({region}) current quota: [bold]{quota['value']}[/bold]\n"
            "Request new quota value? (return to skip):"
        )
        desired_value = IntPrompt.ask(prompt, default=quota["value"])
        if desired_value and desired_value != quota["value"]:
            success, error_message = request_quota_increase(quota_client, region, quota, desired_value=desired_value)

            if not skip_telemetry:
                coiled.add_interaction(
                    action="RequestServiceQuotaIncrease",
                    success=success,
                    quota_name=quota["name"],
                    desired_value=desired_value,
                    current_value=quota["value"],
                    error_message=error_message or None,
                )

            if not success:
                # something failed so stop
                return


def request_quota_increase(
    quota_client, region: str, quota: dict, desired_value: Union[int, float]
) -> Tuple[bool, str]:
    quota_label = quota["class"]
    quota_code = quota["code"]
    try:
        quota_response = quota_client.request_service_quota_increase(
            ServiceCode="ec2",
            QuotaCode=quota_code,
            DesiredValue=float(desired_value),
        )

        request_status = quota_response["RequestedQuota"]["Status"]
        request_case_id = quota_response["RequestedQuota"].get("CaseId")  # might not be in response

        info_text = (
            f"You've submitted a request in increase [bold]{quota_label}[/bold] "
            f"({quota_code}) in {region} to [bold]{desired_value}[/bold]. "
            f"The current status of your request is [bold]{request_status}[/bold]."
        )

        status_text = ""

        if request_status in ("PENDING", "CASE_OPENED"):
            status_text = (
                "Quota requests can take a few days. You can check the status of your request at:\n"
                f"[link]{get_support_case_link(request_case_id)}"
            )
        elif request_status in ("DENIED", "CASE_CLOSED"):
            status_text = (
                "You can find more details about your quota request in the AWS Console:\n"
                f"[link]{get_support_case_link(request_case_id)}"
            )

        if status_text:
            info_text = f"{info_text}\n\n{status_text}"

        print(info_text)

        return True, ""

    except Exception as e:
        error_message = (
            "An error occurred while requesting your quota increase:\n\n"
            f"[red]{e}[/red]\n\n"
            f"You can manually request a quota increase for {quota_label} in the AWS Console at:\n"
            f"[link]{get_quota_link(region, quota_code)}[/link]"
        )
        print(error_message)

        return False, str(e)


def get_support_case_link(case_id: Optional[str]) -> str:
    if case_id:
        return f"https://support.console.aws.amazon.com/support/home#/case/?displayId={case_id}"
    return "https://support.console.aws.amazon.com/support/home#/case/history"


def get_quota_link(region: str, quota_code: str) -> str:
    return f"https://{region}.console.aws.amazon.com/servicequotas/home/services/ec2/quotas/{quota_code}"



import click
from dask import config
from rich.console import Console
from rich.table import Table

from coiled.exceptions import BuildError
from coiled.types import ArchitectureTypesEnum
from coiled.v2.core import setup_logging

from ..core import Cloud, create_software_environment, delete_software_environment
from ..utils import COILED_SERVER
from .utils import CONTEXT_SETTINGS

console = Console()


@click.group(context_settings=CONTEXT_SETTINGS)
def env():
    """Commands for managing Coiled software environments"""
    setup_logging()


@env.command(context_settings=CONTEXT_SETTINGS)
@click.option("-n", "--name", help="Name of software environment, it must be lowercase.")
@click.option("--container", default=None, help="Base docker image to use.")
@click.option(
    "--conda",
    default=None,
    help="Conda environment file.",
    type=click.Path(exists=True),
)
@click.option("--pip", default=None, help="Pip requirements file.", type=click.Path(exists=True))
@click.option(
    "--force-rebuild",
    default=False,
    help="Skip checks for an existing software environment build.",
    is_flag=True,
)
@click.option(
    "--account",
    default=None,
    type=str,
    help="Account to use for creating this software environment.",
)
@click.option(
    "--gpu-enabled",
    is_flag=True,
    show_default=True,
    default=False,
    help="Set CUDA virtual package for Conda",
)
@click.option(
    "--architecture",
    type=click.Choice([e.value for e in ArchitectureTypesEnum]),
    default=ArchitectureTypesEnum.X86_64.value,
    show_default=True,
    help="CPU architecture to use for the software environment",
)
@click.option(
    "--region-name",
    default=None,
    type=str,
    help="AWS or GCP region to use for storing this software environment.",
)
def create(name, container, conda, pip, force_rebuild, account, gpu_enabled, architecture, region_name):
    """Create a Coiled software environment"""
    try:
        create_software_environment(
            name=name,
            container=container,
            conda=conda,
            pip=pip,
            force_rebuild=force_rebuild,
            account=account,
            gpu_enabled=gpu_enabled,
            architecture=architecture,
            region_name=region_name,
        )
    except BuildError as e:
        raise click.ClickException(f"{e}")


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument("name")
def delete(name: str):
    """Delete a Coiled software environment"""
    delete_software_environment(
        name,
    )


@env.command(context_settings=CONTEXT_SETTINGS)
@click.argument("account", default=None, required=False)
def list(account: str):
    """List the Coiled software environments in an account"""
    with Cloud(account=account) as cloud:
        environments = cloud.list_software_environments(
            account,
        )
        account = cloud.default_account
        table = Table(title="Software Environments")
        table.add_column("Name", style="cyan")
        table.add_column("Updated", style="magenta")
        table.add_column("Link", style="magenta")
        table.add_column("Status", style="green")
        server = config.get("coiled.server", COILED_SERVER).replace("8000", "5173")
        account = account or config.get("coiled.account")
        for env_name, env_details in environments.items():
            latest_build = env_details["latest_spec"].get("latest_build")
            if latest_build:
                build_status = latest_build["state"] if latest_build["state"] != "error" else "[red]error[/red]"
                env_url = f"{server}/software/alias/{env_details['id']}/build/{latest_build['id']}?account={account}"
            else:
                build_status = "n/a"
                env_url = f"{server}/software/alias/{env_details['id']}?account={account}"
            table.add_row(
                env_name,
                env_details["updated"],
                env_url,
                build_status,
            )
        console.print(table)


@env.command(
    context_settings=CONTEXT_SETTINGS,
    help="View the details of a Coiled software environment",
)
@click.argument("name")
def inspect(name: str):
    """View the details of a Coiled software environment

    Parameters
    ----------
    name
        Identifier of the software environment to use, in the format (<account>/)<name>. If the software environment
        is owned by the same account as that passed into "account", the (<account>/) prefix is optional.

        For example, suppose your account is "wondercorp", but your friends at "friendlycorp" have an environment
        named "xgboost" that you want to use; you can specify this with "friendlycorp/xgboost". If you simply
        entered "xgboost", this is shorthand for "wondercorp/xgboost".

        The "name" portion of (<account>/)<name> can only contain ASCII letters, hyphens and underscores.

    Examples
    --------
    >>> import coiled
    >>> coiled.inspect("coiled/default")

    """
    with Cloud() as cloud:
        results = cloud.get_software_info(name)
        console.print(results)



import click

from ..utils import CONTEXT_SETTINGS
from .notebook import monitor_sync, start_notebook, stop_notebook


@click.group(name="notebook", context_settings=CONTEXT_SETTINGS)
def notebook():
    """Commands for managing JupyterLab sessions on dedicated VMs in the cloud.

    The Python packages you have installed locally are automatically installed.
    (See https://docs.coiled.io/user_guide/package_sync.html for details.)

    Your local filesystem can also be synced using the `--sync` flag.

    This creates an easy "make my laptop bigger" command, since you'll have
    the same files and environment, just on a bigger machine in the cloud.
    """
    pass


notebook.add_command(start_notebook, "start")
notebook.add_command(stop_notebook, "stop")
notebook.add_command(monitor_sync, "monitor")



import json
import os
import shutil
import subprocess
import sys
import tempfile
import time
from typing import Optional

import click
import dask.config
from rich import print
from rich.panel import Panel
from rich.prompt import Confirm
from rich.table import Table

import coiled

from ...auth import get_local_user
from ..utils import CONTEXT_SETTINGS
from .util import setup_failure

# names to use when tracking actions
ACTION_INIT = "init"
ACTION_CREATE_GAR = "gar:create"
ACTION_BIND_GAR = "gar:policy-binding"
ACTION_BIND_POLICY = "policy-binding"
ACTION_CREATE_ROLE = "roles-create"
ACTION_LIST_KEYS = "keys-list"
ACTION_ENABLE_SERVICE = "services-enable"
ACTION_CREATE_SA = "service-accounts-create"
ACTION_ACTIVATE_SA = "activate-servicer-account"
ACTION_CREATE_KEY = "keys-create"


CONTROL_PLANE_ROLE = """
title: coiled
description: used by Coiled control-plane for creating infrastructure and clusters
stage: GA
includedPermissions:
- bigquery.datasets.create
- bigquery.jobs.create
- bigquery.datasets.get
- bigquery.datasets.update
- bigquery.tables.getData
- compute.acceleratorTypes.list
- compute.addresses.list
- compute.disks.create
- compute.disks.delete
- compute.disks.list
- compute.disks.useReadOnly
- compute.firewalls.create
- compute.firewalls.delete
- compute.firewalls.get
- compute.firewalls.list
- compute.globalOperations.get
- compute.globalOperations.getIamPolicy
- compute.globalOperations.list
- compute.images.create
- compute.images.delete
- compute.images.get
- compute.images.list
- compute.images.setLabels
- compute.images.useReadOnly
- compute.instances.create
- compute.instances.delete
- compute.instances.get
- compute.instances.getSerialPortOutput
- compute.instances.list
- compute.instances.setLabels
- compute.instances.setMetadata
- compute.instances.setServiceAccount
- compute.instances.setTags
- compute.instanceTemplates.create
- compute.instanceTemplates.delete
- compute.instanceTemplates.get
- compute.instanceTemplates.useReadOnly
- compute.machineTypes.get
- compute.machineTypes.list
- compute.networks.create
- compute.networks.delete
- compute.networks.get
- compute.networks.list
- compute.networks.updatePolicy
- compute.projects.get
- compute.projects.setCommonInstanceMetadata
- compute.regionOperations.get
- compute.regionOperations.list
- compute.regions.get
- compute.regions.list
- compute.routers.create
- compute.routers.delete
- compute.routers.get
- compute.routers.list
- compute.routers.update
- compute.routes.delete
- compute.routes.list
- compute.subnetworks.create
- compute.subnetworks.delete
- compute.subnetworks.get
- compute.subnetworks.getIamPolicy
- compute.subnetworks.list
- compute.subnetworks.use
- compute.subnetworks.useExternalIp
- compute.zoneOperations.get
- compute.zoneOperations.list
- compute.zones.list
- iam.serviceAccounts.actAs
- logging.buckets.create
- logging.buckets.get
- logging.buckets.list
- logging.logEntries.create
- logging.logEntries.list
- logging.sinks.create
- logging.sinks.get
- logging.sinks.list
- storage.buckets.create
- storage.buckets.get
- storage.objects.create
- storage.objects.get
- storage.objects.list
- storage.objects.update
"""


DATA_ROLE = """
title: coiled-data
description: service account attached to cluster instances for data access
stage: GA
includedPermissions:
- logging.logEntries.create
- storage.buckets.get
- storage.buckets.create
- storage.objects.create
- storage.objects.get
- storage.objects.list
- storage.objects.update
"""


def has_gcloud():
    return bool(shutil.which("gcloud"))


def gcloud_wrapper(
    command: str,
    show_stdout: bool = False,
    interactive: bool = False,
    error_handler=None,
    action_name: Optional[str] = None,
):
    p = subprocess.run(command.split(" "), capture_output=not interactive)

    stdout = p.stdout.decode(encoding=sys.stdout.encoding or "utf-8") if p.stdout else None
    stderr = p.stderr.decode(encoding=sys.stderr.encoding or "utf-8") if p.stderr else None

    if action_name:
        coiled.add_interaction(
            action=f"gcloud:{action_name}",
            success=p.returncode == 0,
            command=command,
            return_code=p.returncode,
            additional_text=stdout or None,
            error_message=stderr or None,
        )

    if show_stdout:
        print(stdout)

    if p.returncode:
        if stderr and "already exist" in stderr:
            pass

        elif error_handler and error_handler(stderr):
            # error_handler returns True if error was handled
            return False

        else:
            print(f"[red]gcloud returned an error when running command `{command}`:")
            print(Panel(stderr or ""))
            setup_failure(f"gcloud error while running {command}, {stderr}", backend="gcp")
            sys.exit(1)

    return True


def gcloud_enable_service(service: str, project: str):
    command = f"gcloud services enable {service} --project {project}"

    def enable_error_handler(stderr: str):
        if "Billing must be enabled" in stderr:
            print(f"[red]We couldn't enable {service} because of an error related to Google Cloud billing:")
            print(Panel(stderr))
            print("To address this error, please link this project to a Google Cloud billing account at")
            print()
            print(f"[link]https://console.cloud.google.com/billing?project={project}")
            print()
            print("and then re-run [green]coiled setup gcp[/green]")
            # we've handled the error
            setup_failure("Billing not enabled", backend="gcp")
            return True

        return False

    return gcloud_wrapper(
        command,
        error_handler=enable_error_handler,
        action_name=f"{ACTION_ENABLE_SERVICE}:{service}",
    )


def gcloud_make_key(iam_email: str, key_path: str):
    command = f"gcloud iam service-accounts keys create {key_path} --iam-account {iam_email}"
    p = subprocess.run(command.split(" "), capture_output=True)

    # TODO track interaction
    stdout = p.stdout.decode(encoding=sys.stdout.encoding or "utf-8")
    stderr = p.stderr.decode(encoding=sys.stderr.encoding or "utf-8")

    coiled.add_interaction(
        action=f"gcloud:{ACTION_CREATE_KEY}",
        success=p.returncode == 0,
        command=command,
        return_code=p.returncode,
        additional_text=stdout or None,
        error_message=stderr or None,
    )

    if p.returncode:
        if "service-accounts.keys.create" in stderr and "FAILED_PRECONDITION" in stderr:
            print("[red]An error occurred when attempting to create a key for the IAM account[/red]")
            print(stderr)
            print("One possibility is that you already have 10 keys for this account. Existing keys:")
            print()
            gcloud_wrapper(
                f"gcloud iam service-accounts keys list --iam-account={iam_email}",
                show_stdout=True,
                action_name=ACTION_LIST_KEYS,
            )
            print()
            print("You can delete individual keys by running:")
            print()
            print(f"[green]gcloud iam service-accounts keys delete --iam-account={iam_email} [bold]<key_id>[/bold]")
            setup_failure(f"Service account key creation failed {stderr}", backend="gcp")
            return False
        else:
            raise

    return True


def get_gcloud_config(key) -> Optional[str]:
    p = subprocess.run(f"gcloud config get {key}".split(" "), capture_output=True)

    if not p.returncode:
        return p.stdout.decode(encoding=sys.stdout.encoding or "utf-8").strip()
    else:
        print(p.stderr)
        return None


def get_gcloud_json(command):
    p = subprocess.run(command.split(" "), capture_output=True)

    if not p.returncode:
        text = p.stdout.decode(encoding=sys.stdout.encoding or "utf-8").strip()
        return json.loads(text)
    else:
        print(p.stderr)
        return None


def gcloud_check_gar(gar_name: str, region: str) -> bool:
    command = f"gcloud artifacts repositories describe {gar_name} --location {region}"
    p = subprocess.run(command.split(" "), capture_output=True)

    return "format: DOCKER" in p.stdout.decode(encoding=sys.stdout.encoding or "utf-8")


def gcloud_wait_for_repo(region: str, key_path: str) -> bool:
    # it takes a while to give the service account (which we created)
    # access to the artifact registry
    # this function waits (with timeout) until that's ready

    # keep track of currently active account (for local `gcloud`) so we can switch back
    previous_account = get_gcloud_config("account")

    # activate the service account we made --
    # subsequent `gcloud` calls with be using those creds
    gcloud_wrapper(
        f"gcloud auth activate-service-account --key-file {key_path}",
        action_name=ACTION_ACTIVATE_SA,
    )

    success = False
    command = f"gcloud artifacts repositories describe coiled --location {region}"
    t0 = time.time()

    # keep checking if service account has access to artifact registry
    while True:
        p = subprocess.run(command.split(" "), capture_output=True)

        if "format: DOCKER" in p.stdout.decode(encoding=sys.stdout.encoding or "utf-8"):
            success = True
            break
        elif "PERMISSION_DENIED" in p.stderr.decode(encoding=sys.stdout.encoding or "utf-8"):
            if time.time() < t0 + 180:  # wait for up to three minutes
                time.sleep(10)  # 10s delay between retries
                continue
            else:
                break
        else:
            print("[red]An unexpected error occurred while waiting for Google Artifact Registry to be ready.")
            print(Panel(p.stderr.decode(encoding=sys.stdout.encoding or "utf-8")))
            break

    # switch back to the user's account for local `gcloud` commands
    gcloud_wrapper(f"gcloud config set account {previous_account}")
    return success


@click.option(
    "--region",
    default="us-east1",
    help="GCP region to use when setting up your VPC/subnets",
)
@click.option(
    "--enable-gar",
    default=False,
    is_flag=True,
    help="Configure Google Artifact Registry for use with Coiled",
)
@click.option(
    "--manual-final-setup",
    default=False,
    is_flag=True,
    help="Don't automatically send credentials to Coiled, finish setup manually in the web UI",
)
@click.option(
    "--quotas",
    default=False,
    is_flag=True,
    help="Check and potentially request GCP quota increases",
)
@click.option(
    "--quota-link",
    default=False,
    is_flag=True,
    help="Just show link to GCP Console quota page",
)
@click.option(
    "--account",
    default=None,
    help="Coiled account that will be configured. By default, uses your default Coiled account.",
)
@click.option(
    "--export",
    default="",
    type=click.Choice(["", "role", "data-role"]),
    help="Allows you to export role definitions as files",
)
@click.option(
    "--iam-user",
    default="coiled",
    help="Name for role and user to create",
)
@click.option(
    "-y",
    "--yes",
    default=False,
    is_flag=True,
    help="Don't prompt for confirmation, just do it!",
)
@click.option(
    "--use-legacy-setup/--no-use-legacy-setup",
    default=False,
    is_flag=True,
    hidden=True,
    help="Use new setup backend",
)
@click.command(context_settings=CONTEXT_SETTINGS)
def gcp_setup(
    region, enable_gar, manual_final_setup, quotas, quota_link, export, iam_user, yes, account, use_legacy_setup
):
    do_setup(
        region, enable_gar, manual_final_setup, quotas, quota_link, export, iam_user, yes, account, not use_legacy_setup
    )


def do_setup(
    region=None,
    enable_gar=False,
    manual_final_setup=False,
    quotas=False,
    quota_link=False,
    export=None,
    iam_user="coiled",
    yes=False,
    account=None,
    use_shim: bool = False,
) -> bool:
    local_user = get_local_user()

    coiled.add_interaction(
        action="CliSetupGcp",
        success=True,
        local_user=local_user,
        # use keys that match the cli args
        region=region,
        export=export,
        quotas=quotas,
        yes=yes,
    )

    if export:
        export_path = os.path.abspath(f"coiled-{export}.yaml")

        if export == "role":
            with open(export_path, "w") as f:
                f.write(CONTROL_PLANE_ROLE)
            print(f"Exported Coiled role definition to {export_path}")

        elif export == "data-role":
            with open(export_path, "w") as f:
                f.write(DATA_ROLE)
            print(f"Exported Coiled 'data access' role definition to {export_path}")

        return False

    if not has_gcloud():
        print("\n[red]Missing:[/red] The gcloud CLI tool is required for automatic setup.\n")
        if os.path.exists(os.path.join(sys.prefix, "conda-meta")):
            print(
                "Install gcloud with conda using the following command:\n\n"
                "\t[green]conda install -c conda-forge google-cloud-sdk\n"
            )

        print("For more information, see [link]https://docs.coiled.io/user_guide/setup/gcp/cli.html[/link]")

        setup_failure("Gcloud cli missing", backend="gcp")
        return False

    project = get_gcloud_config("project")

    if not project:
        print("No project was set, so you'll need to select or create one...")
        gcloud_wrapper("gcloud init", interactive=True, action_name=ACTION_INIT)

    project = get_gcloud_config("project")
    if not project:
        print("[red]There's still no project set so aborting Coiled setup.")
        setup_failure("No project was set", backend="gcp")
        return False

    region = region or get_gcloud_config("compute/region")

    if not region:
        print(
            "No region is set, you'll need to specify region like so:\n"
            "[green]coiled setup gcp --region us-central1[/green]"
        )
        setup_failure("Region was not set", backend="gcp")
        return False

    if quotas or quota_link:
        show_quotas(project, region, help=not quota_link)
        return False

    base_account_name = iam_user
    base_name_for_role = iam_user.replace("-", "_")  # role can't have `-`

    main_sa = base_account_name
    data_sa = f"{base_account_name}-data"
    main_role = base_name_for_role
    data_role = f"{base_name_for_role}_data"
    gar_name = "coiled"  # this is what control-plane expects

    main_email = f"{main_sa}@{project}.iam.gserviceaccount.com"
    data_email = f"{data_sa}@{project}.iam.gserviceaccount.com"

    resource_description = (
        f"Proposed region for Coiled:\t[green]{region}[/green]\t"
        f"(use `coiled setup gcp --region` to change)\n"
        f"Proposed project for Coiled:\t[green]{project}[/green]\t"
        f"(use `gcloud config set project <project_name>` to set)\n"
        "\n"
        "[bold]The following IAM resources will be created:[/bold]\n"
        "\n"
        f"Service account for creating clusters:\t[green]{main_sa}[/green]\n"
        f"Service account for data access:\t[green]{data_sa}[/green]\n"
        f"IAM Role:\t\t[green]{main_role}[/green] "
        f"(for [green]{main_sa}[/green] service account)\n"
        f"IAM Role:\t\t[green]{data_role}[/green] "
        f"(for [green]{data_sa}[/green] service account)"
    )

    if enable_gar:
        resource_description += "\n" f"Artifact Registry:\t[green]{gar_name}[/green] in region {region}\n"

    print(Panel(resource_description))

    if not yes and not Confirm.ask("Proceed with Google Cloud IAM setup for Coiled?", default=True):
        return False

    with tempfile.TemporaryDirectory() as dir:
        main_file = os.path.join(dir, f"{main_role}.yaml")
        with open(main_file, "w") as f:
            f.write(CONTROL_PLANE_ROLE)

        data_file = os.path.join(dir, f"{data_role}.yaml")
        with open(data_file, "w") as f:
            f.write(DATA_ROLE)

        gcloud_wrapper(
            f"gcloud iam roles create {main_role} --project={project} --file {main_file} --quiet",
            action_name=ACTION_CREATE_ROLE,
        )

        gcloud_wrapper(
            f"gcloud iam roles create {data_role} --project={project} --file {data_file} --quiet",
            action_name=ACTION_CREATE_ROLE,
        )

    gcloud_wrapper(f"gcloud iam service-accounts create {main_sa}", action_name=ACTION_CREATE_SA)
    gcloud_wrapper(f"gcloud iam service-accounts create {data_sa}", action_name=ACTION_CREATE_SA)

    gcloud_wrapper(
        f"gcloud projects add-iam-policy-binding {project} "
        f"--member=serviceAccount:{main_email} "
        f"--role=projects/{project}/roles/{main_role} "
        "--condition=None",
        action_name=ACTION_BIND_POLICY,
    )
    gcloud_wrapper(
        f"gcloud projects add-iam-policy-binding {project} "
        f"--member=serviceAccount:{data_email} "
        f"--role=projects/{project}/roles/{data_role} "
        "--condition=None",
        action_name=ACTION_BIND_POLICY,
    )

    print("Service accounts and roles created.")
    print("Enabling services (this may take a while)...")

    services = [
        "compute",
        "bigquery.googleapis.com",
        "logging",
        "monitoring",
    ]

    if enable_gar:
        services.append("artifactregistry.googleapis.com")

    for service in services:
        print(f"  {service}")
        if not gcloud_enable_service(service, project):
            setup_failure(f"Enabling service {service} failed", backend="gcp")
            return False
    print("Services enabled.")

    if enable_gar:
        print("Setting up Google Artifact Registry (this may take a few minutes)...")
        if not gcloud_check_gar(gar_name, region):
            gcloud_wrapper(
                f"gcloud artifacts repositories create {gar_name} --repository-format=docker --location={region}",
                action_name=ACTION_CREATE_GAR,
            )
        gcloud_wrapper(
            f"gcloud artifacts repositories add-iam-policy-binding {gar_name} "
            "--role=roles/artifactregistry.repoAdmin "
            f"--location={region} "
            f"--member=serviceAccount:{main_email} "
            "--condition=None",
            action_name=ACTION_BIND_GAR,
        )

    show_quotas(project, region)

    if not manual_final_setup:
        print()
        print("[bold]You can setup Coiled to use the Google Cloud credentials you just created.")
        print(
            "The service account key will go to Coiled, where it will be stored "
            "securely and used to create clusters in your Google Cloud account "
            "on your behalf."
        )
        print(
            "This will also create infrastructure in your account like a VPC "
            "and subnets, none of which has a standing cost."
        )
        print()

    if not manual_final_setup and (
        yes
        or Confirm.ask(
            "Setup your Coiled account to use the Google Cloud credentials you just created?",
            default=True,
        )
    ):
        coiled.add_interaction(action="prompt:CoiledSetup", success=True)

        with tempfile.TemporaryDirectory() as dir:
            key_path = os.path.join(dir, "coiled-key.json")
            if not gcloud_make_key(main_email, key_path):
                return False

            if enable_gar:
                print("Waiting for Google Artifact Registry to be ready...")
                if not gcloud_wait_for_repo(region=region, key_path=key_path):
                    # FIXME better instructions about what to do in this case
                    print(
                        "[red]Google Artifact Registry is still not ready. "
                        "You may need to complete account setup manually; "
                        "please contact us if you need help."
                    )
                    setup_failure("Timeout waiting for GAR to be ready", backend="gcp")
                    return False
                print("Google Artifact Registry is ready.")

            print("Setting up Coiled to use your Google Cloud account...")
            coiled.set_backend_options(
                backend="gcp",
                registry_type="gar" if enable_gar else "ecr",
                gcp_region=region,
                gcp_service_creds_file=key_path,
                instance_service_account=data_email,
                use_shim=use_shim,
                account=account,
            )
            coiled.add_interaction(action="CoiledSetup", success=True)

    else:
        coiled.add_interaction(action="prompt:CoiledSetup", success=False)
        with coiled.Cloud() as cloud:
            coiled_account = account or cloud.default_account

        setup_url = (
            f"{dask.config.get('coiled.server', coiled.utils.COILED_SERVER)}/"
            f"{coiled_account}/settings/setup/credentials"
        )

        data_role_console_link = (
            "https://console.cloud.google.com/iam-admin/roles/details/"
            f"projects%3C{project}%3Croles%3C{data_role}?project={project}"
        )

        key_path = os.path.abspath("coiled-key.json")
        if not gcloud_make_key(main_email, key_path):
            return False

        print(
            Panel(
                "You've successfully created service accounts for Coiled "
                "in your Google Cloud project. "
                "You can now complete your Coiled account setup by telling "
                "us how to use these service accounts:\n"
                "\n"
                f"1. Go to [link]{setup_url}[/link] and select GCP.\n"
                "\n"
                f"2. The credential file for [green]{main_sa}[/green] has been saved "
                f"on your local computer at \n"
                f"     {key_path}\n"
                f"   Choose this file for the "
                f"[bold]cluster-creation service account[/bold].\n"
                "\n"
                f"3. Enter [green]{data_email}[/green] as the [bold]data access "
                f"service account[/bold].\n"
                "\n"
                "You'll then have the ability to chose non-default network setup "
                "or container registry settings if desired.\n"
                "\n"
                "[bold]Optional steps[/bold]\n"
                "\n"
                "After you've successfully setup your Coiled account, you'll no "
                "longer need the credential file on your local computer and can "
                "delete this if you wish.\n"
                "\n"
                "We've configured the data access service account with scope for "
                "submitting logs and for accessing Google Storage. If you wish to "
                "add or remove permissions attached to your Coiled clusters for "
                "accessing data, you can do that at \n"
                "\n"
                f"[link]{data_role_console_link}[/link]."
            )
        )

    return True


def show_quotas(project: str, region: str, help: bool = True):
    existing_quotas = get_quotas(project, region)

    quota_types = [
        {
            "name": "CPUs",
            "explanation": "vCPUs per VM * number of VMs",
            "value": existing_quotas.get("CPUS"),
            "unit": "vCPU",
        },
        {
            "name": "Persistent Disk SSD",
            "explanation": "at least 40GB required per VM",
            "value": existing_quotas.get("SSD_TOTAL_GB"),
            "unit": "GB",
        },
        {
            "name": "In-use IP addresses",
            "explanation": "one per VM",
            "value": existing_quotas.get("IN_USE_ADDRESSES"),
            "unit": "addresses",
        },
    ]

    url = (
        "https://console.cloud.google.com/iam-admin/quotas"
        f"?project={project}"
        "&pageState=(%22allQuotasTable%22:"
        "(%22f%22:%22%255B%257B_22k_22_3A_22Quota_22_2C_22t_22_3A10_2C_22v_22_3A_22_5C_22_5C_5C_5C_22"
        "CPUs_5C_5C_5C_22_5C_22_22_2C_22i_22_3A_22displayName_22%257D_2C%257B_22k_22_3A_22_22_2C_22t_"
        f"22_3A10_2C_22v_22_3A_22_5C_22region_3A{region}_5C_22_22_2C_22s_22_3Atrue%257D_2C%257B_22k"
        "_22_3A_22_22_2C_22t_22_3A10_2C_22v_22_3A_22_5C_22OR_5C_22_22_2C_22o_22_3Atrue%257D_2C%257B_"
        "22k_22_3A_22Quota_22_2C_22t_22_3A10_2C_22v_22_3A_22_5C_22"
        "Persistent%2520Disk%2520SSD%2520%2528GB%2529_5C_22_22_2C_22s_22_3Atrue_2C_22i_22_3A_22"
        "displayName_22%257D_2C%257B_22k_22_3A_22_22_2C_22t_22_3A10_2C_22v_22_3A_22_5C_22"
        f"region_3A{region}_5C_22_22_2C_22s_22_3Atrue%257D_2C%257B_22k_22_3A_22_22_2C_22t_22_"
        "3A10_2C_22v_22_3A_22_5C_22OR_5C_22_22_2C_22o_22_3Atrue%257D_2C%257B_22k_22_3A_22"
        "Quota_22_2C_22t_22_3A10_2C_22v_22_3A_22_5C_22In-use%2520IP%2520addresses_5C_22_22_2C_"
        "22s_22_3Atrue_2C_22i_22_3A_22displayName_22%257D_2C%257B_22k_22_3A_22_22_2C_22t_22_3A10_"
        f"2C_22v_22_3A_22_5C_22region_3A{region}_5C_22_22_2C_22s_22_3Atrue%257D%255D%22,%22s%22:%5B"
        "(%22i%22:%22displayName%22,%22s%22:%220%22),"
        "(%22i%22:%22effectiveLimit%22,%22s%22:%221%22),"
        "(%22i%22:%22currentPercent%22,%22s%22:%221%22),"
        "(%22i%22:%22currentUsage%22,%22s%22:%221%22),"
        "(%22i%22:%22serviceTitle%22,%22s%22:%220%22),"
        "(%22i%22:%22displayDimensions%22,%22s%22:%220%22)%5D))"
    )
    if help:
        table = Table(title=f"Existing quotas for [bold]{project}[/bold] in [bold]{region}[/bold]")
        table.add_column("Quota type", justify="left")
        table.add_column("Existing value", justify="right")
        table.add_column("Cluster will use...", justify="left", overflow="fold")

        for quota_type in quota_types:
            if quota_type["value"]:
                quota_value = f"[bold]{int(quota_type['value'])}[/bold] {quota_type['unit']}"
            else:
                quota_value = "(could not retrieve)"

            table.add_row(quota_type["name"], quota_value, quota_type["explanation"])

        print(
            "Coiled runs in your own Google Cloud account, so your cluster sizes will be "
            "constrained by your GCP quotas.\n"
        )

        print(table)

        print(
            Panel(
                "These quotas are set per-region, so you'll need a high enough quota in the region where you'll "
                "be running your Coiled clusters.\n"
                "\n"
                "The default Coiled instance type ([bold]e2-standard-4[/bold]) has [bold]4 vCPUs[/bold], "
                "so a 10 worker cluster (plus scheduler) would require [bold]44 vCPUs[/bold], "
                "[bold]440 GB[/bold] Persistent Disk SSD, and [bold]11 IP addresses[/bold].\n"
                "\n"
                f"For [bold]{project}[/bold] (project) in [bold]{region}[/bold] "
                f"you can view the current quotas and request increases in the [link={url}]Google Cloud Console[/link]."
            )
        )
    else:
        print(url)


def get_quotas(project: str, region: str) -> dict:
    quota_by_metric = {}
    quotas = get_gcloud_json(f"gcloud compute regions describe {region} --project {project} --format json")
    if quotas:
        quota_by_metric = {quota["metric"]: quota["limit"] for quota in quotas["quotas"]}
    return quota_by_metric



import asyncio
import hashlib
import json
import os
import platform
import re
import subprocess
import sys
import typing
from base64 import urlsafe_b64encode
from collections import defaultdict
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional, Union, cast
from urllib.parse import urlparse

from importlib_metadata import Distribution, PackagePath, PathDistribution
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version
from rich.progress import Progress
from typing_extensions import Literal

from coiled.pypi_conda_map import CONDA_TO_PYPI
from coiled.types import CondaPackage, CondaPlaceHolder, PackageInfo
from coiled.utils import COILED_LOCAL_PACKAGE_PREFIX, parse_file_uri, recurse_importable_python_files

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()


class ResilientDistribution(PathDistribution):
    """Subclass of Distribution that adds more resilient methods for retrieving files"""

    def _read_files_egginfo_installed(self):
        """
        Read installed-files.txt and return lines in a similar
        CSV-parsable format as RECORD: each file should be placed
        relative to the site-packages directory and must be
        quoted (since file names can contain literal commas).

        This file is written when the package is installed by pip,
        but it might not be written for other installation methods.
        Assume the file is accurate if it exists.
        """
        text = self.read_text("installed-files.txt")
        # Prepend the .egg-info/ subdir to the lines in this file.
        # But this subdir is only available from PathDistribution's
        # self._path.
        subdir = self._path
        if not text or not subdir:
            return

        site_pkgs_path = cast(Path, self.locate_file("")).resolve()
        for name in text.splitlines():
            # relpath will add .. to a path to make it relative to site-packages,
            # so use that instead of Path.relative_to (which will raise an error)
            path = Path(os.path.relpath((subdir / name).resolve(), site_pkgs_path))
            yield f'"{path.as_posix()}"'


def convert_conda_to_pypi_name(name: str):
    return CONDA_TO_PYPI.get(name, name)


def normalize_version(version: str):
    """Attempt to normalize version between conda and pip without tripping over conda not supporting PEP440"""
    # Bail on non-PEP440 versions like 2023c, which will get parsed as 2023rc0
    if not len(version.split(".")) > 2:
        return version
    else:
        try:
            # Normalize things like 23.04.00 to 23.4.0
            return str(parse_version(version))
        # Fallback to original version if its unparseable like 1.7.1dev.rapidsai23.04
        except InvalidVersion:
            return version


async def scan_conda(prefix: Path, progress: Optional[Progress] = None) -> typing.Dict[str, List[PackageInfo]]:
    conda_meta = prefix / "conda-meta"
    if conda_meta.exists() and conda_meta.is_dir():
        conda_packages = [
            CondaPackage(json.load(metafile.open("r")), prefix=prefix)
            for metafile in conda_meta.iterdir()
            if metafile.suffix == ".json"
        ]
        packages: List[PackageInfo] = []
        if progress:
            for task in progress.track(
                asyncio.as_completed([handle_conda_package(pkg) for pkg in conda_packages]),
                description=f"Scanning {len(conda_packages)} conda packages",
                total=len(conda_packages),
            ):
                r = await task
                if r:
                    packages.append(r)
        else:
            packages = [
                pkg for pkg in await asyncio.gather(*[handle_conda_package(pkg) for pkg in conda_packages]) if pkg
            ]
        # it's possible for multiple similar packages to be "installed"
        # eg importlib-metadata & importlib_metadata
        # we have to check later which one is actually being imported
        result: Dict[str, List[PackageInfo]] = defaultdict(list)
        for pkg in packages:
            result[pkg["name"]].append(pkg)
        return result
    else:
        return {}


async def handle_conda_package(pkg: CondaPackage) -> Optional[PackageInfo]:
    # Are there conda packages that install multiple python packages?
    metadata_location = next(
        (pkg.prefix / Path(fp).parent for fp in pkg.files if fp.endswith(("METADATA", "PKG-INFO"))), None
    )
    if metadata_location:
        if not metadata_location.exists():
            # a file for this package no longer exists
            # likely pip installed a new version
            # removing the conda installed version
            return None
        else:
            dist = ResilientDistribution(pkg.prefix / metadata_location)  # type: ignore
            name = dist.metadata["Name"] or pkg.name
            path = Path(str(dist._path))
    else:
        name = pkg.name
        path = (
            # Just pick a file so we can know this isn't a metapackage
            Path(pkg.prefix / pkg.files[0]) if pkg.files else None
        )
    return {
        "channel": pkg.channel,
        "path": path,
        "channel_url": pkg.channel_url,
        "source": "conda",
        "conda_name": pkg.name,
        "subdir": pkg.subdir,
        "name": convert_conda_to_pypi_name(name),
        "version": pkg.version,
        "wheel_target": None,
    }


async def handle_dist(dist: Distribution, locations: List[Path]) -> Optional[Union[PackageInfo, CondaPlaceHolder]]:
    # Sometimes the dist name is blank (seemingly only on Windows?)
    if not dist.name:
        return
    installer = dist.read_text("INSTALLER") or ""
    installer = installer.rstrip()
    # dist._path can sometimes be a zipp.Path or something else
    dist_path = Path(str(dist._path))  # type: ignore
    if installer == "conda":
        return CondaPlaceHolder(name=convert_conda_to_pypi_name(dist.name), path=dist_path)
    elif dist_path.parent.suffix == ".egg":
        # egg files can be a directory OR a zip file
        # the zipp implementation of Path always uses
        # linux style seperators so we strip them too
        return {
            "name": dist.name,
            "path": dist_path.parent,
            "source": "pip",
            "channel": None,
            "subdir": None,
            "channel_url": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": str(dist_path.parent).rstrip(os.sep + "/"),
        }
    else:
        direct_url_metadata = dist.read_text("direct_url.json")
        if direct_url_metadata:
            url_metadata = json.loads(direct_url_metadata)
            url = url_metadata.get("url")
            if not url:
                # no url in this file
                # invalid PEP-610 so don't do anything
                pass
            elif url_metadata.get("vcs_info"):
                # PEP-610 Source is VCS
                vcs_info = url_metadata.get("vcs_info")
                vcs: Literal["git", "hg", "bzr", "svn"] = vcs_info["vcs"]
                commit = vcs_info["commit_id"]
                url = url_metadata["url"]
                pip_url = f"{vcs}+{url}@{commit}"
                return {
                    "name": dist.name,
                    "path": dist_path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": pip_url,
                }
            elif str((Path("pypoetry") / "artifacts")) in url_metadata["url"]:
                # if the install source is actually the pre 1.2 poetry cache location
                # they this is actually just normal a pypi
                # and we can ignore direct_url.json
                pass
            elif url_metadata.get("archive_info") is not None:
                # PEP-610 - Source is an archive/wheel, somewhere!
                p = urlparse(url)
                if p.scheme == "file":
                    url = str(parse_file_uri(url))
                return {
                    "name": dist.name,
                    "path": dist_path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": url,
                }
            elif url_metadata.get("dir_info") is not None:
                # PEP-610 - Source is a local directory
                path = parse_file_uri(url)
                return {
                    "name": dist.name,
                    "path": path,
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": str(path),
                }
        egg_links = []
        for location in locations:
            egg_link_pth = location / Path(dist.name).with_suffix(".egg-link")
            if egg_link_pth.is_file():
                egg_links.append(location / Path(dist.name).with_suffix(".egg-link"))
        if egg_links:
            return {
                "name": dist.name,
                "path": dist_path.parent,
                "source": "pip",
                "channel": None,
                "channel_url": None,
                "subdir": None,
                "conda_name": None,
                "version": dist.version,
                "wheel_target": str(dist_path.parent),
            }
        return {
            "name": dist.name,
            "path": dist_path,
            "source": "pip",
            "channel": None,
            "channel_url": None,
            "subdir": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": None,
        }


def _is_hash_match(dist: Distribution, pkg_paths: Dict[str, PackagePath], path: str):
    dist_path = Path(str(dist._path)).parent  # type: ignore
    pkg_path = pkg_paths.get(path)
    if pkg_path is not None and pkg_path.hash is not None:  # type: ignore
        pkg_hash = pkg_path.hash  # type: ignore
        hash_func = getattr(hashlib, pkg_hash.mode)
        absolute_path = dist_path / pkg_path
        if absolute_path.exists() and absolute_path.is_file():
            with absolute_path.open("rb") as f:
                actual_hash = urlsafe_b64encode(hash_func(f.read()).digest()).strip(b"=").decode()
                if actual_hash == pkg_hash.value:
                    return True
    return False


async def scan_pip(
    locations: List[Path], progress: Optional[Progress] = None
) -> typing.Dict[str, Union[PackageInfo, CondaPlaceHolder]]:
    # distributions returns ALL distributions
    # even ones that are not active
    # this is a trick so we only get the distribution
    # that is last in stack
    locations = [location for location in locations if location.exists() and location.is_dir()]
    paths: List[str] = [str(location) for location in locations]
    encoding = sys.stdout.encoding or "utf-8"
    for location in locations:
        for fp in location.iterdir():
            if fp.suffix in [".pth", ".egg-link"]:
                try:
                    file_text = fp.read_text()
                except UnicodeDecodeError:
                    try:
                        file_text = fp.read_text(encoding=encoding)
                    except UnicodeDecodeError:
                        logger.debug("Could not read file %s with encoding %s", fp, encoding, exc_info=True)
                        continue

                for line in file_text.split("\n"):
                    if line.startswith("#"):
                        continue
                    elif line.startswith(("import", "import\t")):
                        continue
                    elif line.rstrip() == ".":
                        continue
                    else:
                        p = location / Path(line.rstrip())
                        full_path = str(p.resolve())
                        if p.exists() and full_path not in paths:
                            paths.append(full_path)
    # can't use ResilientDistribution here properly without monkey patching it
    dists: List[Distribution] = [dist for dist in Distribution.discover(path=list(paths))]
    packages = []
    if progress:
        for task in progress.track(
            asyncio.as_completed([handle_dist(dist, locations) for dist in dists]),
            total=len(dists),
            description=f"Scanning {len(dists)} python packages",
        ):
            packages.append(await task)
    else:
        packages = await asyncio.gather(*(handle_dist(dist, locations) for dist in dists))

    # Resolve duplicate packages
    pkgs_by_name = {}
    for pkg in packages:
        if pkg:
            pkg_name = pkg["name"]
            # For duplicate .dist-info directories, we need to check which
            # version is actually importable
            existing_pkg = pkgs_by_name.get(pkg_name)
            if existing_pkg is None:
                pkgs_by_name[pkg_name] = pkg
            else:
                # Compare hashes to actual files
                new_dist = ResilientDistribution(pkg["path"])
                old_dist = ResilientDistribution(existing_pkg["path"])
                new_dist_path = new_dist._path
                old_dist_path = old_dist._path
                new_is_egg_info = new_dist_path.name.endswith(".egg-info")  # type: ignore
                old_is_egg_info = old_dist_path.name.endswith(".egg-info")  # type: ignore
                new_is_dist_info = new_dist_path.name.endswith(".dist-info")  # type: ignore
                old_is_dist_info = old_dist_path.name.endswith(".dist-info")  # type: ignore
                if (new_is_egg_info and not old_is_egg_info) or (new_is_dist_info and not old_is_dist_info):
                    continue
                if (not new_is_egg_info and old_is_egg_info) or (not new_is_dist_info and old_is_dist_info):
                    pkgs_by_name[pkg_name] = pkg
                    continue

                if new_is_egg_info and old_is_egg_info:
                    # This should never happen
                    logger.debug(
                        "Found two egg-info directories with the same name: %s and %s", new_dist_path, old_dist_path
                    )

                new_pkg_paths = {
                    str(f): f
                    for f in (new_dist.files or [])
                    if (
                        not f.name.endswith(".pyc")
                        and f.parent.name != new_dist_path.name  # type: ignore
                        and f.hash is not None  # type: ignore
                    )
                }
                old_pkg_paths = {
                    str(f): f
                    for f in (old_dist.files or [])
                    if (
                        not f.name.endswith(".pyc")
                        and f.parent.name != old_dist_path.name  # type: ignore
                        and f.hash is not None  # type: ignore
                    )
                }
                old_paths = set(old_pkg_paths.keys())
                new_paths = set(new_pkg_paths.keys())
                same_hashes = {
                    path
                    for path in old_paths.intersection(new_paths)
                    if old_pkg_paths[path].hash.mode == new_pkg_paths[path].hash.mode  # type: ignore
                    and old_pkg_paths[path].hash.value == new_pkg_paths[path].hash.value  # type: ignore
                }
                paths_to_check = new_paths.union(old_paths) - same_hashes

                for path in paths_to_check:
                    # Since we are only checking files that have different hashes,
                    # we can just assume the new version is correct on the first
                    # match.
                    if _is_hash_match(new_dist, new_pkg_paths, path):
                        pkgs_by_name[pkg_name] = pkg
                        break
                    if not _is_hash_match(old_dist, old_pkg_paths, path):
                        logger.debug("Encountered path that does not match either version: %s", path)

    return pkgs_by_name


async def scan_prefix(
    prefix: Optional[Path] = None,
    base_prefix: Optional[Path] = None,
    progress: Optional[Progress] = None,
    locations: Optional[List[Path]] = None,
) -> typing.List[PackageInfo]:
    # TODO: private conda channels
    # TODO: detect pre-releases and only set --pre flag for those packages (for conda)

    if not prefix:
        prefix = Path(sys.prefix).resolve()
    if not base_prefix:
        base_prefix = Path(sys.base_prefix).resolve()
    if not locations:
        locations = [Path(p).resolve() for p in sys.path]
        cwd = Path.cwd().resolve()
        if cwd not in locations:
            locations.insert(0, cwd)
    conda_env_future = asyncio.create_task(scan_conda(prefix=prefix, progress=progress))
    # only pass locations to support testing, otherwise we should be using sys.path
    pip_env_future = asyncio.create_task(scan_pip(locations=locations, progress=progress))
    conda_env = await conda_env_future
    pip_env = await pip_env_future
    filtered_conda = {}
    # the pip list is the "truth" of what is imported for python deps
    for pypi_name, packages in conda_env.items():
        # if a package exists in the pip list but is not a conda place holder
        # then the conda package wont be imported and should be discarded
        found = False
        if pip_env.get(pypi_name):
            found = True
            pip_package = pip_env[pypi_name]
            if isinstance(pip_package, CondaPlaceHolder):
                del pip_env[pypi_name]
                # find the conda package that actually matches with what is importable
                importable_package = next(
                    (p for p in packages if p["path"]),
                    None,
                )
                if importable_package is None:
                    logger.warning("Could not find importable conda package for %s", pypi_name)
                else:
                    filtered_conda[pypi_name] = importable_package
            else:
                matching_package = next(
                    (
                        p
                        for p in packages
                        if normalize_version(p["version"]) == normalize_version(pip_package["version"])
                    ),
                    None,
                )
                if matching_package:
                    # if the versions match, we can fall back to using the conda version
                    del pip_env[pypi_name]
                    filtered_conda[pypi_name] = matching_package
        if not found:
            # a non python package and safe to include
            filtered_conda[pypi_name] = packages[0]
    # remove conda placeholders (there won't be any by here but this makes pyright happy)
    pip_env = {pkg_name: pkg for pkg_name, pkg in pip_env.items() if not isinstance(pkg, CondaPlaceHolder)}
    results = sorted(
        list(pip_env.values()) + list(filtered_conda.values()),
        key=lambda pkg: pkg["name"],
    )
    # get set of urls for all packages that were installed via pip
    pkg_urls = set()
    for pkg in results:
        if pkg["wheel_target"]:
            url = cast(str, pkg["wheel_target"])
            if "http://" in url or "https://" in url:
                url = re.sub(r"^(git|hg|svn|bzr)\+", "", url)
                url = re.sub(r"@\w+$", "", url)
                url = url.rstrip("/")
                pkg_urls.add(url)

    # Handle modules that are not installed via pip or conda
    pkg_paths = {pkg["path"].resolve() for pkg in results if pkg["path"]}
    extra_paths = {
        p
        for p in locations
        if p not in (prefix, base_prefix) and prefix not in p.parents and base_prefix not in p.parents
    } - pkg_paths
    for extra_path in extra_paths:
        if not extra_path.is_dir():
            continue
        if any(recurse_importable_python_files(extra_path)):
            # Skip directories that are the same as a package that was installed via pip
            git_dir = extra_path / ".git"
            if git_dir.exists():
                try:
                    encoding = sys.stdout.encoding or "utf-8"
                    remote_output = subprocess.check_output(
                        ["git", "remote", "-v"],
                        cwd=extra_path,
                        encoding=encoding,
                    )
                    remote_urls = set()
                    for line in remote_output.splitlines():
                        url = line.split()[1].strip()
                        if url.startswith("git@"):
                            url = "https://" + url[4:].replace(":", "/")
                        if url.endswith(".git"):
                            url = url[:-4]
                        url = url.rstrip("/")
                        remote_urls.add(url)

                except subprocess.CalledProcessError:
                    remote_urls = {}
                if pkg_urls.intersection(remote_urls):
                    continue

            cleaned_name = re.sub(r"[^\w\d.]+", "_", extra_path.name, re.UNICODE)
            if cleaned_name.startswith("_"):
                cleaned_name = cleaned_name.lstrip("_")
            results.append(
                {
                    "name": COILED_LOCAL_PACKAGE_PREFIX + cleaned_name,
                    "path": extra_path,
                    "source": "pip",
                    "version": "0.0.0",
                    "channel_url": None,
                    "channel": None,
                    "subdir": None,
                    "conda_name": None,
                    "wheel_target": str(extra_path),
                }
            )
    return sorted(results, key=lambda pkg: pkg["name"])



# This file was auto-generated by scripts/update_pypi_conda_mapping.py
# DO NOT MANUALLY EDIT
PYPI_TO_CONDA = {
    "Boruta": "boruta_py",
    "Flask-Cors": "flask_cors",
    "Flask-JSON": "flask_json",
    "SoundFile": "pysoundfile",
    "aiohttp_cors": "aiohttp-cors",
    "annoy": "python-annoy",
    "art": "ascii-art",
    "asana": "python-asana",
    "backtrace": "python-backtrace",
    "bash_completion": "py-bash-completion",
    "bilby-pipe": "bilby_pipe",
    "blis": "cython-blis",
    "captest": "pvcaptest",
    "cdo": "python-cdo",
    "click_config_file": "click-config-file",
    "cockroachdb": "cockroachdb-python",
    "conda_lock": "conda-lock",
    "conda_mirror": "conda-mirror",
    "confluent-kafka": "python-confluent-kafka",
    "coreapi": "python-coreapi",
    "coreschema": "python-coreschema",
    "cryptography_vectors": "cryptography-vectors",
    "cufflinks": "python-cufflinks",
    "dask": "dask-core",
    "data-morph-ai": "data-morph-ai",
    "dataclasses": "dataclasses",
    "datadotworld": "datadotworld-py",
    "datalad_container": "datalad-container",
    "delegator.py": "delegator",
    "dials_data": "dials-data",
    "docker": "docker-py",
    "dye_score": "dye-score",
    "eccodes": "python-eccodes",
    "empyrical_dist": "empyrical-dist",
    "enum34": "enum34",
    "esprima": "esprima-python",
    "et_xmlfile": "et-xmlfile",
    "evalml": "evaml-core",
    "exceptiongroup": "exceptiongroup",
    "extract-msg": "msg-extractor",
    "fastjsonschema": "python-fastjsonschema",
    "flair": "python-flair",
    "flake8_nb": "flake8-nb",
    "flatbuffers": "python-flatbuffers",
    "flex": "flex-swagger",
    "flit_core": "flit-core",
    "gnupg": "gnupg-py",
    "google": "googlesearch",
    "google-cloud-bigquery": "google-cloud-bigquery-core",
    "graphviz": "python-graphviz",
    "gssapi": "python-gssapi",
    "h2o": "h2o-py",
    "hdfs": "python-hdfs",
    "hjson": "hjson-py",
    "htcondor": "python-htcondor",
    "iam_units": "iam-units",
    "ib_insync": "ib-insync",
    "import_metadata": "import_metadata",
    "importlib_metadata": "importlib-metadata",
    "installer": "python-installer",
    "intake-astro": "intake-accumulo",
    "intake_geopandas": "intake-geopandas",
    "ioos_tools": "ioos-tools",
    "jsii": "python-jsii",
    "jupyter-client": "jupyter_client",
    "jupyter_jaeger": "jupyter-jaeger",
    "jupyterlab_git": "jupyterlab-git",
    "jupyterlab_latex": "jupyterlab-latex",
    "kaleido": "python-kaleido",
    "kubernetes": "python-kubernetes",
    "libaio": "python-libaio",
    "matplotlib": "matplotlib-base",
    "md_toc": "md-toc",
    "message_ix": "message-ix",
    "msgpack": "msgpack-python",
    "mss": "python-mss",
    "nbconvert_utils": "nbconvert-utils",
    "ndarray_listener": "ndarray-listener",
    "neo": "python-neo",
    "neo4j": "neo4j-python-driver",
    "nest_asyncio": "nest-asyncio",
    "node-semver": "python-node-semver",
    "nvidia-ml-py3": "nvidia-ml",
    "opencv-python": "opencv",
    "ortools": "ortools-python",
    "p-astro": "p_astro",
    "pandas_flavor": "pandas-flavor",
    "paragraph": "python-paragraph",
    "prince": "prince-factor-analysis",
    "pvlib": "pvlib-python",
    "pyct": "pyct-core",
    "pyqt4": "pyqt",
    "pyqt5": "pyqt",
    "pytest_check_links": "pytest-check-links",
    "python-datamatrix": "datamatrix",
    "python-fileinspector": "fileinspector",
    "python-opencv": "opencv",
    "python-pseudorandom": "pseudorandom",
    "python-pygaze": "pygaze",
    "python-qdatamatrix": "qdatamatrix",
    "python-qnotifications": "qnotifications",
    "pywin32": "pywin32-on-windows",
    "quantum-grove": "grove",
    "radio_beam": "radio-beam",
    "redis": "redis-py",
    "rocketpyalpha": "rocketpy",
    "sage_flatsurf": "sage-flatsurf",
    "sagemaker_inference": "sagemaker-inference-toolkit",
    "sagemaker_mxnet_training": "sagemaker_mxnet_container",
    "scikit-build": "scikit-build",
    "scikit-learn": "scikit-learn",
    "setuptools_scm": "setuptools-scm",
    "sounddevice": "python-sounddevice",
    "spherical-functions": "spherical_functions",
    "sphinx-rtd-theme": "sphinx_rtd_theme",
    "stjudecloud-oliver": "oliver",
    "symengine": "python-symengine",
    "tables": "pytables",
    "termstyle": "python-termstyle",
    "torch": "pytorch",
    "torch_geometric": "pytorch_geometric",
    "trino": "trino-python-client",
    "typing": "typing",
    "useDAVE": "dave",
    "vsts": "vsts-python-api",
    "xxhash": "python-xxhash",
}

CONDA_TO_PYPI = {
    "aiohttp-cors": "aiohttp_cors",
    "ascii-art": "art",
    "bilby_pipe": "bilby-pipe",
    "boruta_py": "Boruta",
    "click-config-file": "click_config_file",
    "cockroachdb-python": "cockroachdb",
    "conda-lock": "conda_lock",
    "conda-mirror": "conda_mirror",
    "cryptography-vectors": "cryptography_vectors",
    "cython-blis": "blis",
    "dask-core": "dask",
    "data-morph-ai": "data-morph-ai",
    "dataclasses": "dataclasses",
    "datadotworld-py": "datadotworld",
    "datalad-container": "datalad_container",
    "datamatrix": "python-datamatrix",
    "dave": "useDAVE",
    "delegator": "delegator.py",
    "dials-data": "dials_data",
    "docker-py": "docker",
    "dye-score": "dye_score",
    "empyrical-dist": "empyrical_dist",
    "enum34": "enum34",
    "esprima-python": "esprima",
    "et-xmlfile": "et_xmlfile",
    "evaml-core": "evalml",
    "exceptiongroup": "exceptiongroup",
    "fileinspector": "python-fileinspector",
    "flake8-nb": "flake8_nb",
    "flask_cors": "Flask-Cors",
    "flask_json": "Flask-JSON",
    "flex-swagger": "flex",
    "flit-core": "flit_core",
    "gnupg-py": "gnupg",
    "google-cloud-bigquery-core": "google-cloud-bigquery",
    "googlesearch": "google",
    "grove": "quantum-grove",
    "h2o-py": "h2o",
    "hjson-py": "hjson",
    "iam-units": "iam_units",
    "ib-insync": "ib_insync",
    "import_metadata": "import_metadata",
    "importlib-metadata": "importlib_metadata",
    "intake-accumulo": "intake-astro",
    "intake-geopandas": "intake_geopandas",
    "ioos-tools": "ioos_tools",
    "jupyter-jaeger": "jupyter_jaeger",
    "jupyter_client": "jupyter-client",
    "jupyterlab-git": "jupyterlab_git",
    "jupyterlab-latex": "jupyterlab_latex",
    "matplotlib-base": "matplotlib",
    "md-toc": "md_toc",
    "message-ix": "message_ix",
    "msg-extractor": "extract-msg",
    "msgpack-python": "msgpack",
    "nbconvert-utils": "nbconvert_utils",
    "ndarray-listener": "ndarray_listener",
    "neo4j-python-driver": "neo4j",
    "nest-asyncio": "nest_asyncio",
    "nvidia-ml": "nvidia-ml-py3",
    "oliver": "stjudecloud-oliver",
    "opencv": "python-opencv",
    "ortools-python": "ortools",
    "p_astro": "p-astro",
    "pandas-flavor": "pandas_flavor",
    "prince-factor-analysis": "prince",
    "pseudorandom": "python-pseudorandom",
    "pvcaptest": "captest",
    "pvlib-python": "pvlib",
    "py-bash-completion": "bash_completion",
    "pyct-core": "pyct",
    "pygaze": "python-pygaze",
    "pyqt": "pyqt5",
    "pysoundfile": "SoundFile",
    "pytables": "tables",
    "pytest-check-links": "pytest_check_links",
    "python-annoy": "annoy",
    "python-asana": "asana",
    "python-backtrace": "backtrace",
    "python-cdo": "cdo",
    "python-confluent-kafka": "confluent-kafka",
    "python-coreapi": "coreapi",
    "python-coreschema": "coreschema",
    "python-cufflinks": "cufflinks",
    "python-eccodes": "eccodes",
    "python-fastjsonschema": "fastjsonschema",
    "python-flair": "flair",
    "python-flatbuffers": "flatbuffers",
    "python-graphviz": "graphviz",
    "python-gssapi": "gssapi",
    "python-hdfs": "hdfs",
    "python-htcondor": "htcondor",
    "python-installer": "installer",
    "python-jsii": "jsii",
    "python-kaleido": "kaleido",
    "python-kubernetes": "kubernetes",
    "python-libaio": "libaio",
    "python-mss": "mss",
    "python-neo": "neo",
    "python-node-semver": "node-semver",
    "python-paragraph": "paragraph",
    "python-sounddevice": "sounddevice",
    "python-symengine": "symengine",
    "python-termstyle": "termstyle",
    "python-xxhash": "xxhash",
    "pytorch": "torch",
    "pytorch_geometric": "torch_geometric",
    "pywin32-on-windows": "pywin32",
    "qdatamatrix": "python-qdatamatrix",
    "qnotifications": "python-qnotifications",
    "radio-beam": "radio_beam",
    "redis-py": "redis",
    "rocketpy": "rocketpyalpha",
    "sage-flatsurf": "sage_flatsurf",
    "sagemaker-inference-toolkit": "sagemaker_inference",
    "sagemaker_mxnet_container": "sagemaker_mxnet_training",
    "scikit-build": "scikit-build",
    "scikit-learn": "scikit-learn",
    "setuptools-scm": "setuptools_scm",
    "spherical_functions": "spherical-functions",
    "sphinx_rtd_theme": "sphinx-rtd-theme",
    "trino-python-client": "trino",
    "typing": "typing",
    "vsts-python-api": "vsts",
}



import os

try:
    from prefect.client.schemas import FlowRun  # type: ignore
    from prefect.infrastructure.container import Optional, anyio  # type: ignore
    from prefect.workers.process import (  # type: ignore
        ProcessJobConfiguration,
        ProcessWorker,
    )
except ImportError:
    raise ImportError("`prefect` must be installed to use Coiled prefect extension module.")


class CoiledJobConfiguration(ProcessJobConfiguration):
    """CoiledJobConfiguration"""

    # TODO: Implement relavant fields


class CoiledWorker(ProcessWorker):
    """
    A PrefectWorker specific to Coiled.

    Can run flows on Coiled infrastructure either in combination with CoiledTaskRunner to
    map flow tasks to dask tasks, as done with ``DaskTaskRunner`` but will handle mapping
    flows to the existing cluster this worker is running on.

    Alternatively, can be used without ``CoiledTaskRunner`` if you are wanting to run flows
    normally on a single VM.
    """

    type = "coiled-worker"
    _block_type_name = "Coiled Worker"
    _block_type_slug = "coiled-worker"

    async def run(
        self,
        flow_run: FlowRun,
        configuration: CoiledJobConfiguration,
        task_status: Optional[anyio.abc.TaskStatus] = None,
    ):
        # Propagate cluster name to flow environment, which CoiledTaskRunner will pick up
        # to use as its cluster (which is the same this Worker is running on)
        envvar = "COILED__CLUSTER_NAME"
        if envvar in os.environ:
            configuration.env[envvar] = os.environ[envvar]

        return await super().run(flow_run, configuration, task_status)



import os
import pathlib
import subprocess
import sys
import tempfile
import threading
import time
from shutil import copytree
from typing import Optional, Union

import click
from distributed import Client, Scheduler, SchedulerPlugin, UploadDirectory, Worker, WorkerPlugin
from rich import print

import coiled
from coiled.utils import logger

from .utils import CONTEXT_SETTINGS


@click.group(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def prefect(ctx):
    """Prefect interface"""


@click.group(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def worker(ctx):
    """Worker interface"""


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--prefect-workspace",
    help="Name of the Prefect workspace.",
    required=True,
)
@click.option(
    "--prefect-key",
    help="API key for Prefect workspace, used to login Prefect workers.",
    default="",
)
@click.option(
    "--prefect-pool",
    default="default",
    help="Name of pool for Prefect worker.",
)
@click.option(
    "--prefect-work-queue",
    default="default",
    help="Name of Prefect work-queue for worker.",
)
@click.option(
    "--cluster-name",
    default=None,
    help="Name of Coiled cluster",
)
@click.option(
    "--n-workers",
    default=1,
    type=int,
    help="Number of Prefect workers to launch",
)
@click.option("--disable-adaptive", default=False, flag_value=True, help="Opt out of defaulting to an adaptive cluster")
@click.option(
    "--adaptive-min",
    default=0,
    type=int,
    help="Minimum workers in adaptive cluster configuration",
)
@click.option(
    "--adaptive-max",
    default=None,
    type=int,
    help="Maximum workers in adaptive cluster configuration",
)
@click.option(
    "--container",
    default=None,
    type=str,
    help="Docker image uri to use, ex: coiled/default:sha-2021.06.0",
)
def start(
    cluster_name: Optional[str],
    n_workers: int,
    prefect_workspace: str,
    prefect_work_queue: str,
    prefect_pool: str,
    disable_adaptive: bool,
    adaptive_min: int,
    adaptive_max: int,
    prefect_key: str = "",
    software: Optional[str] = None,
    container: Optional[str] = None,
):
    """
    Deploy Prefect Worker(s) on a Coiled cluster.

    This should be ran from the same code (and environment) used when
    deploying your Prefect flow/deployment. It will sync that directory's code
    to the VM(s) where the worker(s) will be running. Running the command from
    the desired Python environment also ensures all packages are synced to the
    Worker(s)/VM(s).
    """
    if not prefect_key:
        fn = os.path.expanduser(os.path.join("~", ".prefect", "profiles.toml"))
        try:
            import toml

            data = toml.load(fn)
            prefect_key = data["profiles"][data["active"]]["PREFECT_API_KEY"]
        except Exception:
            print(
                "[red]Missing API Key:[/red] Please log in to Prefect with\n\n"
                "    [green]prefect cloud login[/green]\n\n"
                "Or specify your Prefect API key with `--prefect-key <<KEY>>`"
            )
            sys.exit(1)
    cluster = coiled.Cluster(
        name=cluster_name,
        n_workers=n_workers - 1,  # One prefect worker on scheduler
        shutdown_on_close=False,
        scheduler_options={"idle_timeout": None},
        container=container,
        tags={"coiled-cluster-type": "prefect"},
    )
    client = cluster.get_client()

    _sync_local_storage(client)

    prefect_plugin = PrefectWorkerPlugin(
        cluster_name=cluster.name,  # type: ignore
        workspace=prefect_workspace,
        key=prefect_key,
        work_queue=prefect_work_queue,
        pool=prefect_pool,
    )
    client.register_worker_plugin(prefect_plugin, name="prefect-worker")
    client.register_scheduler_plugin(prefect_plugin, name="prefect-worker")

    if not disable_adaptive:
        adaptive_plugin = AdaptiveSchedulerPlugin(cluster.name, minimum=adaptive_min, maximum=adaptive_max)
        client.register_scheduler_plugin(adaptive_plugin)

    logger.info("Finished.")


worker.add_command(start, "start")
prefect.add_command(worker, "worker")


class AdaptiveSchedulerPlugin(SchedulerPlugin):
    """
    To persist cluster adaptivity, the ``Adaptive`` obj
    needs to stay alive. This does simply that, ``kwargs``
    is passed to ``coiled.Cluster.adapt`` and the result is
    stored in this plugin.
    """

    def __init__(self, name, **kwargs):
        self.kwargs = kwargs.copy()
        self.name = name

    async def start(self, scheduler) -> None:
        """Run when a new client connects"""
        if not hasattr(self, "adapt"):
            cluster = await coiled.Cluster(self.name, asynchronous=True)
            self.adapt = cluster.adapt(**self.kwargs)


class PrefectWorkerPlugin(SchedulerPlugin, WorkerPlugin):
    """
    A Scheduler/WorkerPlugin to run a Prefect worker on
    a Dask scheduler/worker as a subprocess.
    """

    proc: subprocess.Popen

    def __init__(
        self, cluster_name: str, workspace: str, key: str, work_queue: Optional[str] = None, pool: str = "default"
    ):
        self.cluster_name = cluster_name
        self.workspace = workspace
        self.key = key
        self.work_queue = work_queue
        self.pool = pool

    def start_prefect_worker(self, vm: Union[Worker, Scheduler]):
        subprocess.check_output(f"prefect cloud login -w {self.workspace} -k {self.key}".split())

        os.putenv("COILED__CLUSTER_NAME", self.cluster_name)
        os.putenv("PREFECT_EXTRA_ENTRYPOINTS", "coiled.extensions.prefect")

        cmd = "prefect worker start --type coiled-worker"
        cmd += f" --name {vm.name}" if hasattr(vm, "name") else ""  # type: ignore
        cmd += f" --pool {self.pool}" if self.pool else ""
        cmd += f" --work-queue {self.work_queue}" if self.work_queue else ""

        self.proc = subprocess.Popen(cmd.split())

        rc = self.proc.poll()
        if rc is not None:
            raise RuntimeError(f"Failed to start prefect worker, exit code: {rc}")

        # Monitor prefect worker doesn't die w/o us knowning about it.
        self._lock = threading.Lock()
        self._monitor_thread = threading.Thread(target=self.monitor_prefect_worker)
        self._monitor_thread.start()

    def stop_prefect_worker(self):
        with self._lock:
            self.proc.terminate()
            self.proc.wait()

    def monitor_prefect_worker(self):
        while True:
            with self._lock:
                rc = self.proc.poll()
            if rc is not None:
                if rc == 0:
                    break
                raise RuntimeError(f"Prefect worker died, exit code: {rc}")
            time.sleep(10)

    # SchedulerPlugin bridge
    async def start(self, scheduler: Scheduler):
        self.start_prefect_worker(scheduler)

    async def close(self):
        self.stop_prefect_worker()

    # WorkerPlugin bridge
    def setup(self, worker: Worker):
        self.start_prefect_worker(worker)

    def teardown(self, worker: Worker):
        self.stop_prefect_worker()


# TODO: PR to distributed to support UploadDirectory plugin for scheduler
# A proper implementation requires distributed>=2023.5.1 which is higher
# than prefect-dask's latest (0.2.4) allows: distributed<=2023.3.1,>=2022.5.0
# Ref: https://github.com/PrefectHQ/prefect-dask/blob/70a93dc6154e36782ea4b81151ce51ec447ceb1a/requirements.txt#L3
class UploadDirectorySchedulerPlugin(UploadDirectory, SchedulerPlugin):
    async def start(self, scheduler: Scheduler):
        # Remove when prefect-dask can be used w/ distributed>=2023.5.1
        scheduler.local_directory = "/tmp"  # type: ignore
        await self.setup(scheduler)


def _sync_local_storage(client: Client):
    logger.info("Syncing local storage to VMs...")
    # TODO: Hacky hack! - We're assuming LocalStorage for Prefect deployment here.
    #       To make the VMs look the same, we need to modify the output of UploadDirectory to
    #       the same path of the locally deployed workflow, assumed to be the cwd in this context.
    #       So we inject a file to echo the output location then copy files to same path on the VM
    #       as was found locally. Ideally, we could pass a destination dir to UploadDirectory.
    workflow_dir = pathlib.Path.cwd()  # Maybe CLI option, assumed running in same dir as workflow
    with tempfile.TemporaryDirectory() as tmpdir:
        copytree(str(workflow_dir), str(tmpdir), dirs_exist_ok=True)
        f = pathlib.Path(tmpdir).joinpath("__discover_location.py")
        f.write_text(
            """
import pathlib

def location():
    return pathlib.Path(__file__).parent
            """
        )
        upload_dir_plugin = UploadDirectory(str(tmpdir), restart=True, update_path=True)
        client.register_worker_plugin(upload_dir_plugin)

        upload_dir_plugin = UploadDirectorySchedulerPlugin(str(tmpdir), restart=True, update_path=True)
        client.register_scheduler_plugin(upload_dir_plugin)

    def copy_to_expected():
        """Running on workers, copy inital UploadDirectory of workflow code to same local path"""
        from __discover_location import location  # type: ignore

        # Copy Prefect workflow over to path that matches the local path it was deployed from
        # to match the LocalStorage
        copytree(str(location()), str(pathlib.Path(workflow_dir)), dirs_exist_ok=True)

    client.run(copy_to_expected)
    client.run_on_scheduler(copy_to_expected)



# Get function backing "coiled install" command
# Get function backing "coiled env inspect" command
from .cli.env import inspect as _inspect
from .core import (
    add_interaction,
    create_api_token,
    create_software_environment,
    delete_software_environment,
    diagnostics,
    get_billing_activity,
    get_notifications,
    get_software_info,
    list_api_tokens,
    list_core_usage,
    list_gpu_types,
    list_instance_types,
    list_local_versions,
    list_performance_reports,
    list_software_environments,
    list_user_information,
    performance_report,
    revoke_all_api_tokens,
    revoke_api_token,
    set_backend_options,
)
from .v2 import (
    AWSOptions,
    BackendOptions,
    CloudV2 as Cloud,
    Cluster,
    FirewallOptions,
    GCPOptions,
    better_cluster_logs,
    cluster_logs,
    create_cluster,
    delete_cluster,
    list_clusters,
)
from .function import run, function
from importlib_metadata import version

inspect = _inspect.callback
del _inspect

# Register coiled configuration values with Dask's config system
from . import config

del config

# Top-level coiled.config attribute


def __getattr__(name):
    if name == "config":
        import dask.config

        return dask.config.get("coiled")
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


__version__ = version("coiled")

__all__ = [
    "__version__",
    "Cloud",
    "Cluster",
    "add_interaction",
    "better_cluster_logs",
    "cluster_logs",
    "create_cluster",
    "create_software_environment",
    "delete_cluster",
    "delete_software_environment",
    "diagnostics",
    "get_billing_activity",
    "get_notifications",
    "get_software_info",
    "list_clusters",
    "list_core_usage",
    "list_gpu_types",
    "list_instance_types",
    "list_local_versions",
    "list_performance_reports",
    "list_software_environments",
    "list_user_information",
    "performance_report",
    "set_backend_options",
    "list_api_tokens",
    "revoke_api_token",
    "revoke_all_api_tokens",
    "create_api_token",
    "AWSOptions",
    "GCPOptions",
    "BackendOptions",
    "FirewallOptions",
]



import asyncio

import click

from ..utils import login_if_required
from .utils import CONTEXT_SETTINGS


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-s", "--server", help="Coiled server to use", hidden=True)
@click.option("-t", "--token", multiple=True, help="Coiled user token")
@click.option("-a", "--account", help="Coiled account")
@click.option(
    "--retry/--no-retry",
    default=True,
    help="Whether or not to automatically ask for a new token if an invalid token is entered",
)
@click.option("--browser/--no-browser", default=True, help="Open browser with page where you grant access")
def login(server, token, account, retry, browser):
    """Configure your Coiled account credentials"""
    # allow token split across multiple --token args, so we can have shorter lines for cloudshell command
    token = "".join(token) if token else None
    asyncio.run(
        login_if_required(
            server=server, token=token, account=account, save=True, use_config=False, retry=retry, browser=browser
        )
    )



import click

import coiled


def find_cluster(cloud: coiled.Cloud, cluster: str):
    if cluster and cluster.isnumeric():
        cluster_id = int(cluster)
    elif cluster:
        # get cluster by name
        try:
            clusters = cloud.get_clusters_by_name(name=cluster)
            if clusters:
                recent_cluster = clusters[-1]
            else:
                raise click.ClickException(
                    f"Unable to find cluster with name '{cluster}' in account '{cloud.default_account}'"
                )

            cluster_id = recent_cluster["id"]

        except coiled.errors.DoesNotExist:
            cluster_id = None
    else:
        # default to most recent cluster
        clusters = cloud.list_clusters(max_pages=1)
        if not clusters:
            raise ValueError(f"Unable to find any clusters for account '{cloud.default_account}'")
        match = max(clusters, key=lambda c: c["id"])
        cluster_id = match["id"]

    if not cluster_id:
        raise click.ClickException(f"Unable to find cluster `{cluster}`")

    try:
        cluster_info = cloud.cluster_details(cluster_id)
    except coiled.errors.ServerError as e:
        if "not found" in str(e).lower():
            raise click.ClickException(f"Unable to find cluster with id {cluster_id}")
        else:
            raise e

    return cluster_info
